/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.persistence.Access;
import javax.persistence.AccessType;
import javax.persistence.Column;
import javax.persistence.MapKey;
import javax.persistence.OrderBy;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import ws.ladder.annotation.Computed;
import ws.ladder.annotation.Initialize;
import ws.ladder.model.event.CollectionEvent;
import ws.ladder.model.event.PropertyEvent;
import ws.ladder.service.DataAccessService;
import ws.ladder.util.ClassUtil;
import ws.ladder.util.I18NUtils;
import ws.ladder.util.ObjectCreator;


public abstract class AbstractProperty implements ExtendedProperty {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());

	protected AccessType accessType;
	protected boolean    needsInitialization;
	protected boolean    computed;

	// The following fields are used to set/retrieve value
	protected Method     getterMethod;    
	protected Method     setterMethod;   	
	protected Method     customGetterMethod;    
	protected Method     customSetterMethod;  	
	protected Method     customAddMethod;    
	protected Method     customRemoveMethod;
	protected Map<String, Method> prePropertyLogicMethods; // methods by tag
	protected Field      field;

	protected ExtendedProperty positionProperty;
	protected ExtendedProperty mapKeyOf;
	protected ExtendedProperty indexOf;	
	private String       mapPath;
	private Property     mappedBy;
	private Property     mapOf;
	private Type         type;
	private Type         parentType;

	// Collection related
	protected Type       keyType;
	protected Type       elementType;	

	private List<String> aliasNames = new ArrayList<String>();

	public AbstractProperty(Type type, Type parentType) {
		this.type = type;
		this.parentType = parentType;
	}

	@Override
	public Type getType() {
		return type;
	}

	public void setType(Type type) {
		this.type = type;
	}

	@Override
	public List<?> getAliasNames() {
		return aliasNames;
	}	

	@Override
	public Type getKeyType() {
		return keyType;
	}

	@Override
	public Type getElementType() {
		return elementType;
	}	

	public static boolean isCascadable(Property property) {
		if(property == null)
			return true;
		return property.isContainment();
	}

	public static boolean isRequired(Property property) {
		return property != null && !property.isNullable();	
	}	

	public static boolean isCollection(Property property) {
		return property != null && property.isMany();
	}	

	@Override
	public Property getOpposite() {
		if(getMappedBy() != null)
			return getMappedBy();

		if(getMapOf() != null)
			return getMapOf();

		return null;
	}

	@Override
	public Property getMappedBy() {
		return mappedBy;
	}

	public void setMappedBy(Property property, String mapPath) {
		this.mappedBy = property;

		if(property != null) {
			((ExtendedProperty)mappedBy).setMapPath(mapPath);
			((AbstractProperty)property).setMapOf(this);
		}
	}

	@Override
	public Property getMapOf() {
		return mapOf;
	}	

	public void setMapOf(Property property) {
		this.mapOf = property;
	}	

	@Override
	public Type getContainingType() {
		return parentType;
	}	

	protected Annotation getAnnotation(Class annotationClass) {
		if(getterMethod != null && getterMethod.isAnnotationPresent(annotationClass))
			return getterMethod.getAnnotation(annotationClass);
		else if(field != null && field.isAnnotationPresent(annotationClass))
			return field.getAnnotation(annotationClass);

		return null;
	}	

	public Method getGetterMethod() {
		return getterMethod;
	}

	protected void init() {
		Class<?> instanceClass = getContainingType().getInstanceClass();

		getterMethod = ClassUtil.getGetterMethod(instanceClass, getName());
		logger.debug("Class name: " + instanceClass.getName() + ", property name: " + getName());

		if(getterMethod == null) // No point in continuing as the field is not present
			return;

		Access access = getterMethod.getAnnotation(Access.class);
		if(access != null)
			accessType = access.value();
		else {
			field = ClassUtil.getField(instanceClass, getName());
			if(field == null && getterMethod != null) // try the method name
				field = ClassUtil.getField(instanceClass, getterMethod.getName());

			if(field != null) {
				access = field.getAnnotation(Access.class);
				if(access != null)
					accessType = access.value();
			} else
				logger.warn("Unable to get Field object. The field name is not the same as getter property name " + getName() + ", skipping field access check.");
		}

		setterMethod = ClassUtil.getSetterMethod(instanceClass, getName());

		if(getterMethod == null || field == null)
			logger.warn("getter or field name is not accessible: " + getName());

		if( (getterMethod != null && getterMethod.isAnnotationPresent(Initialize.class)) || 
				(field != null && field.isAnnotationPresent(Initialize.class)) 
				)
			needsInitialization = true;

		if( (getterMethod != null && getterMethod.isAnnotationPresent(Computed.class)) || 
				(field != null && field.isAnnotationPresent(Computed.class)) 
				)
			computed = true;		
		
		// set custom setters or getters
		if(ClassUtil.getCustomGetter(instanceClass, getName()) != null)
			customGetterMethod = ClassUtil.getCustomGetter(instanceClass, getName());
		if(ClassUtil.getCustomSetter(instanceClass, getName()) != null)
			customSetterMethod = ClassUtil.getCustomSetter(instanceClass, getName());
		if(ClassUtil.getCustomAdd(instanceClass, getName()) != null)
			customAddMethod = ClassUtil.getCustomAdd(instanceClass, getName());
		if(ClassUtil.getCustomRemove(instanceClass, getName()) != null)
			customRemoveMethod = ClassUtil.getCustomRemove(instanceClass, getName());
		prePropertyLogicMethods = ClassUtil.getPrePropertyLogic(this);

		initColumnName();
	}
	
	public Method getPrePropertyLogic() {
		return prePropertyLogicMethods.get(ClassUtil.EMPTY_TAG);
	}
	
	public Method getPrePropertyLogic(String tag) {
		return prePropertyLogicMethods.get( (tag == null) ? ClassUtil.EMPTY_TAG : tag );
	}

	/**
	 * Used in JPA annotated persistence stores, to map the index column
	 */
	private void initColumnName() {
		Column column = (Column) getAnnotation(Column.class);
		if(column != null)
			aliasNames.add(column.name());
	}

	@Override
	public void setMapKeyOf(ExtendedProperty property) {
		this.mapKeyOf = property;
	}

	@Override
	public void setIndexOf(ExtendedProperty property) {
		this.indexOf = property;
	}	

	@Override
	public void initPositionProperty() {

		MapKey mapKey = (MapKey) getAnnotation(MapKey.class);
		if(mapKey != null && (mapKey.name() == null || "".equals(mapKey.name())) ) {// return the elements identifier value 
			positionProperty = (ExtendedProperty) ((ExtendedType)getElementType()).getIdentifierProperty();
			return;
		}

		OrderBy orderBy = (OrderBy) getAnnotation(OrderBy.class);
		if(orderBy != null && (orderBy.value() == null || "".equals(orderBy.value())) ) {
			positionProperty = (ExtendedProperty) ((ExtendedType)getElementType()).getIdentifierProperty();
			return;
		}		

		if(getElementType() != null) {
			if(mapKey != null) {
				positionProperty = (ExtendedProperty) ((ExtendedType)getElementType()).getProperty(mapKey.name());
				if(positionProperty != null)
					positionProperty.setMapKeyOf(this);
			}

			if(orderBy != null) {
				String orderByPropertyName = orderBy.value().split("\\s+")[0];
				positionProperty = (ExtendedProperty) ((ExtendedType)getElementType()).getProperty(orderByPropertyName);
				logger.debug("initPositionProperty: orderBy is not null => " + getName() + " , orderByPropertyName: " + orderByPropertyName + " isPositionProperty: " + (positionProperty != null));				
				if(positionProperty != null) {
					positionProperty.setIndexOf(this);
					if(getKeyType() == null)
						keyType = positionProperty.getType();
				}
			}
		}
	}	

	@Override
	public Property getPositionProperty() {
		return positionProperty;
	}

	public Method getCustomSetter() {
		return customSetterMethod;
	}

	public Method getCustomAdd() {
		return customAddMethod;
	}	

	public Method getCustomRemove() {
		return customRemoveMethod;
	}	

	public abstract void init(DataAccessService das);

	@Override
	public AccessType getAccessType() {
		if(accessType == null)
			return ((ExtendedType)getContainingType()).getAccessType();
		else
			return accessType;
	}

	@Override
	public boolean isContainment() {
		return false;
	}

	@Override
	public Object getValue(Object dataObject) 
	{	
		if(field == null && getterMethod == null) {
			// The object does not have the desired property
			String[] params = new String[2];
			params[0] = getName();
			params[1] = getType().getInstanceClass().getName();
			logger.warn(I18NUtils.getResource(  "exception.propertyNotFound",I18NUtils.CORE_RESOURCES, params));			
			return null;
		} else {		
			return query(dataObject);
		}		
	}

	@Override
	public void setValue(Object dataObject, Object propertyValue) 
	{	
		if(field == null && getterMethod == null) {
			// This field is not present, could be a VO class generation issue if the copy is between JPA and VO
			String[] params = new String[2];
			params[0] = getName();
			params[1] = getType().getInstanceClass().getName();
			logger.warn(I18NUtils.getResource(  "exception.propertyNotFound",I18NUtils.CORE_RESOURCES, params));
			return;
		} else {
			executeUpdate(dataObject, propertyValue);
		}
	}	

	public void setCustomValue(Object dataObject, PropertyEvent event) 
	{
		Object instance = ClassUtil.getInstance(dataObject);

		Object[] args = {event};
		try {
			ClassUtil.invokeMethodAsPrivileged(instance, customSetterMethod, args);
		} catch (Exception e) {
			throw ClassUtil.wrapRun(e);
		}
	}	

	@Override
	public void setCustomAdd(Object dataObject, CollectionEvent event) 
	{
		Object instance = ClassUtil.getInstance(dataObject);

		Object[] args = {event};
		try {
			ClassUtil.invokeMethodAsPrivileged(instance, customAddMethod, args);
		} catch (Exception e) {
			throw ClassUtil.wrapRun(e);
		}
	}		

	@Override
	public void setCustomRemove(Object dataObject, CollectionEvent event) 
	{
		Object instance = ClassUtil.getInstance(dataObject);

		Object[] args = {event};
		try {
			ClassUtil.invokeMethodAsPrivileged(instance, customRemoveMethod, args);
		} catch (Exception e) {
			throw ClassUtil.wrapRun(e);
		}
	}	

	private Object query(Object dataObject) 
	{
		Object instance = ClassUtil.getInstance(dataObject);

		// Set the value in the field  
		Object[] args = {};
		try {
			if(AccessType.PROPERTY.equals(getAccessType())) {
				try {
					return ClassUtil.invokeMethodAsPrivileged(instance, getterMethod, args);
				} catch (Exception e) { // fallback to field access
					logger.warn("Falling back to field access for method : " + getterMethod.getName());
					return ClassUtil.invokeFieldAsPrivileged(instance, field, null, true);
				}
			} else {
				try {
					return ClassUtil.invokeFieldAsPrivileged(instance, field, null, true);
				} catch (Exception e) { // fallback to method access
					logger.warn("Falling back to method access for field : " + field.getName());					
					return ClassUtil.invokeMethodAsPrivileged(instance, getterMethod, args);	
				}

			} 
		} catch (Exception e) {
			throw ClassUtil.wrapRun(e);
		}
	}	

	private <T> void executeUpdate(Object dataObject, Object propertyValue) 
	{
		// We need to work with the instance objects and not the DataObject objects
		Object instance = ClassUtil.getInstance(dataObject);
		propertyValue = ClassUtil.getInstance(propertyValue);

		// Set the value in the field  
		Object[] args = {propertyValue};
		try {
			if(AccessType.PROPERTY.equals(getAccessType())) {
				try {
					ClassUtil.invokeMethodAsPrivileged(instance, setterMethod, args);
				} catch (Exception e) { // fallback to field access
					logger.warn("Falling back to field access for method : " + setterMethod.getName());					
					ClassUtil.invokeFieldAsPrivileged(instance, field, propertyValue, false);
				}
			} else {
				try {
					ClassUtil.invokeFieldAsPrivileged(instance, field, propertyValue, false);
				} catch (Exception e) { // fallback to method access
					logger.warn("Falling back to method access for field : " + field.getName());						
					ClassUtil.invokeMethodAsPrivileged(instance, setterMethod, args);	
				}
			} 
		} catch (Exception e) {
			throw ClassUtil.wrapRun(e);
		}
	}

	public Object shallowCopy(Object dataObject, ObjectCreator creator) {

		try {
			Object propertyValue = getValue(dataObject);

			// TODO: iterate the properties and copy them
		} catch (Exception e) {
			throw ClassUtil.wrapRun(e);
		}
		return null;
	}

	@Override
	public boolean needsInitialization() {
		return needsInitialization;
	}
	
	@Override
	public boolean isComputed() {
		return computed;
	}

	@Override
	public boolean isTransient() {	
		return (field != null) ? Modifier.isTransient(field.getModifiers()) : Modifier.isTransient(getterMethod.getModifiers());

	}	

	@Override
	public boolean isBiDirectional() {
		return getOpposite() != null;
	}

	@Override
	public boolean isSymmetricalBiDirectionalType() {
		if(!isBiDirectional())
			return false;

		ExtendedProperty opposite = (ExtendedProperty) getOpposite();

		if(!isMany() && !opposite.isMany()) 
			return this.type == opposite.getType();

		if(isMany()) {
			if(opposite.isMany()) {
				return getElementType() == opposite.getElementType();
			} else {
				return getElementType() == opposite.getType();
			}
		} else {
			if(opposite.isMany()) {
				return this.type == opposite.getElementType();
			} else {
				return this.type == opposite.getType();
			}
		}
	}

	public void setMapPath(String mapPath) {
		this.mapPath = mapPath;
	}

	@Override
	public String getMapPath() {
		return this.mapPath;
	}

	@Override
	public boolean isDataType() {
		if(isMany()) {
			if(getElementType() != null)
				return getElementType().isDataType();
			else
				return getType().isDataType();
		} else
			return getType().isDataType();
	}

	@Override
	public void linkBackPointer(ExtendedDataObject dataObject) {
		ExtendedProperty backPointer = (ExtendedProperty) getOpposite();
		if(backPointer == null)
			return;

		ExtendedDataObject otherSide = (ExtendedDataObject) dataObject.getDataObject(this);
		if(otherSide == null)
			return;

		if(isMany()) {
			for(ExtendedDataObject otherSideElement: otherSide.getList()) {
				if(backPointer.isMany()) {
					if(backPointer.isSet()) // Only Set is supported
						((Set) backPointer.getValue(otherSideElement)).add(dataObject.getInstance());
				} else
					backPointer.setValue(otherSideElement, dataObject.getInstance());
			}
		} else {
			if(backPointer.isMany()) {
				if(backPointer.isSet()) // Only Set is supported
					((Set) backPointer.getValue(otherSide)).add(dataObject.getInstance());
			} else
				backPointer.setValue(otherSide, dataObject.getInstance());
		}

	}

	@Override
	public void unlinkBackPointer(ExtendedDataObject dataObject) {
		ExtendedProperty backPointer = (ExtendedProperty) getOpposite();
		if(backPointer == null)
			return;

		ExtendedDataObject otherSide = (ExtendedDataObject) dataObject.getDataObject(this);
		if(otherSide == null)
			return;

		if(isMany()) {
			for(ExtendedDataObject otherSideElement: otherSide.getList()) {
				if(backPointer.isMany()) {
					if(backPointer.isSet()) // Only Set is supported
						((Set) backPointer.getValue(otherSideElement)).remove(dataObject.getInstance());
				} else
					backPointer.setValue(otherSideElement, null);
			}
		} else {
			if(backPointer.isMany()) {
				if(backPointer.isSet()) // Only Set is supported
					((Set) backPointer.getValue(otherSide)).remove(dataObject.getInstance());
			} else
				backPointer.setValue(otherSide, null);
		}
	}

	public abstract void initMappedBy(DataAccessService das);
}
