/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.persistence.Entity;

import ws.ladder.annotation.DataModel;
import ws.ladder.service.DataAccessService;
import ws.ladder.util.ClassUtil;

public abstract class AbstractType implements ExtendedType {
	private TypeMapper              typeMapper;   // Used to get the derivedClass from the referenceClass and vice versa
	private ClassResolver           classResolver;
	private boolean                 immutable;
	protected Map<String, Property> properties;
	protected ExtendedType          rootEntityType;
	private   Set<ExtendedType>     subTypes;	

	private static final HashSet<Class<?>> WRAPPER_TYPES = getWrapperTypes();
	private static final HashSet<Class<?>> BASIC_TYPES = getBasicTypes();

	public AbstractType() {
		classResolver = new ClassResolver(this);
	}

	public ExtendedType getRootEntityType() {
		return rootEntityType;
	}
	
	@Override
	public void defineSubtypes(List<ExtendedType> entityTypes) {
		subTypes = new HashSet<ExtendedType>();
		
		for(ExtendedType type: entityTypes) {
			if(this.getInstanceClass().isAssignableFrom(type.getInstanceClass()))
				subTypes.add(type);
		}
	}
	
	@Override
	public Set<ExtendedType> getSubtypes() {
		return subTypes;
	}

	public void initRootEntityType(DataAccessService das) {
		Class<?> rootEntityClass = getInstanceClass();

		Class<?> potentialRoot = rootEntityClass.getSuperclass();		
		while(potentialRoot != null) {
			Annotation annotation = ClassUtil.getClassAnnotation(potentialRoot, Entity.class);

			if(annotation != null && annotation.annotationType() == Entity.class) {
				Entity entity = (Entity) annotation;
				if(entity.name() != null && !"".equals(entity.name()))
					rootEntityClass = potentialRoot;
			}

			potentialRoot = potentialRoot.getSuperclass();
		}

		rootEntityType = (ExtendedType) das.getType(rootEntityClass);
	}

	protected static boolean isWrapperType(Class<?> clazz)
	{
		return WRAPPER_TYPES.contains(clazz);
	}

	private static HashSet<Class<?>> getWrapperTypes()
	{
		HashSet<Class<?>> ret = new HashSet<Class<?>>();
		ret.add(Boolean.class);
		ret.add(Character.class);
		ret.add(Byte.class);
		ret.add(Short.class);
		ret.add(Integer.class);
		ret.add(Long.class);
		ret.add(Float.class);
		ret.add(Double.class);
		ret.add(Void.class);
		return ret;
	}

	private static HashSet<Class<?>> getBasicTypes()
	{
		HashSet<Class<?>> ret = new HashSet<Class<?>>();
		ret.add(String.class);
		ret.add(java.math.BigDecimal.class);
		ret.add(java.sql.Date.class);
		ret.add(java.sql.Time.class);
		ret.add(java.sql.Timestamp.class);
		ret.add(java.util.Date.class);
		return ret;
	}    

	protected void initImmutable() {
		Annotation annotation = ClassUtil.getClassAnnotation(getInstanceClass(), DataModel.class);
		if(annotation != null && annotation.annotationType() == DataModel.class) {
			boolean value = ((DataModel)annotation).immutable();
			if(value)
				this.immutable = value;
		}	
	}	

	public boolean isImmutable() {
		return immutable;
	}

	public void setImmutable(boolean value) {
		this.immutable = value;
	}

	public static boolean isBasicType(Class<?> clazz) {
		return clazz.isPrimitive() || isWrapperType(clazz) || BASIC_TYPES.contains(clazz);
	}

	@Override
	public boolean isReferenceType() {
		return true;
	}

	@Override
	public ExtendedType getReferenceType() {
		return this;
	}

	@Override
	public TypeMapper getTypeMapper() {
		return typeMapper;
	}

	public void setTypeMapper(TypeMapper typeMapper) {
		this.typeMapper = typeMapper;
	}


	@Override
	public boolean isDataType() {
		return properties.size() == 0;
	}

	@Override
	public ClassResolver getClassResolver() {
		return classResolver;
	}

	public void setClassResolver(ClassResolver classResolver) {
		this.classResolver = classResolver;
	}	

	@Override
	public List<Property> getProperties() {
		return sort(new ArrayList<Property>(properties.values()));
	}

	@Override
	public Property getProperty(String path) {
		int delim = path.indexOf(Settings.PATH_DELIMITER);
		if(delim == -1)
			return properties.get(path);
		else {
			Property property = getProperty(path.substring(0, delim)); 
			Type propertyType = property.getType();
			if(property.isMany())
				propertyType = ((ExtendedProperty)property).getElementType();
			return propertyType.getProperty(path.substring(delim+1));
		}		
	}	
	
	@Override
	public Property getPropertyByAlias(String name) {
		for(Property property: properties.values()) {
			if(property.getAliasNames().contains(name))
				return property;
		}
		
		return null;
	}
	
	@Override
	public void initPositionProperty() {
		for(Property property: properties.values())
			((ExtendedProperty)property).initPositionProperty();
	}	

	@Override
	public List<Type> getEmbeddableTypes() {
		return new ArrayList<Type>();
	}

	/**
	 * Move all the collections to the end. This does not order between collections. So if there is a dependency from one collection to another that needs to be resolved using postProcess
	 * e.g., a custom annotation
	 */	
	protected List<Property> sort(List<Property> properties) {
		List<Property> dataTypes = new ArrayList<Property>();
		List<Property> toMany = new ArrayList<Property>();
		List<Property> toOne = new ArrayList<Property>();

		for(Property property: properties) {
			if(property.getType().isDataType()) {
				dataTypes.add(property);
			} else {
				if(property.isMany())
					if(property.isContainment())
						toMany.add(property);
					else
						toMany.add(0, property);
				else
					if(property.isContainment())
						toOne.add(property);
					else
						toOne.add(0, property);
			}
		}
		List<Property> result = dataTypes;
		result.addAll(toOne);
		result.addAll(toMany);

		return result;
	}	

	@Override
	public void invokePostLogic(Object instance) {
		Set<Method> postLogicMethods = ClassUtil.getPostLogicMethods(this.getInstanceClass());
		for(Method method: postLogicMethods) {
			try {
				Object[] args = {};
				ClassUtil.invokeMethodAsPrivileged(instance, method, args);
			} catch (Exception e) {
				throw ClassUtil.wrapRun(e);
			}
		}
	}
	
	@Override
	public Property getUserKey() {
		String userKeyString = ClassUtil.getUserKey(this);
		return (userKeyString != null) ? getProperty(userKeyString) : null;		
	}

	@Override
	public Property getCollectionUserKey() {
		String collUserKeyString = ClassUtil.getCollectionUserKey(this);
		return (collUserKeyString != null) ? getProperty(collUserKeyString) : null;
	}	
}
