/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model;


import ws.ladder.model.operation.Operation;
import ws.ladder.util.ObjectCreator;

public class CallInfo {

	private Object           input;
	private Object           output;
	private ExtendedProperty inputProperty;
	private CallInfo         parent;
	private Settings         settings;	
	private Operation        operation;
	private ObjectCreator    inputObjectCreator; // Ensure we are not sharing objects between the two object creators 	
	private ObjectCreator    outputObjectCreator; // needed to create the root target data object
	private ProcessingStage  stage;

	public Object getInput() {
		return input;
	}

	public void setInput(Object from) {
		this.input = from;
		validate();
	}

	public Object getOutput() {
		return output;
	}

	public void setOutput(Object to) {
		this.output = to;
		validate();		
	}

	public ExtendedProperty getInputProperty() {
		return inputProperty;
	}

	public void setInputProperty(ExtendedProperty property) {
		this.inputProperty = property;
	}

	public ExtendedDataObject getParentOutputEntity() {
		CallInfo parent = getParent();
		ExtendedDataObject result = (ExtendedDataObject) parent.getOutput();
		while( ((ExtendedType)result.getType()).isEmbedded() ) {
			parent = parent.getParent();
			result = (ExtendedDataObject) parent.getOutput();
		}

		return result;
	}

	public CallInfo getParent() {
		return parent;
	}

	public void setParent(CallInfo parent) {
		this.parent = parent;
	}

	public Settings getSettings() {
		return settings;
	}

	public void setSettings(Settings settings) {
		this.settings = settings;
	}

	public CallInfo(Object to, CallInfo parent, ExtendedProperty property) {
		this(null, to, parent, property);

		if(!isDataType()) {
			ExtendedDataObject parentSource = (ExtendedDataObject) getParent().getInput();
			ExtendedDataObject sourceRootDO = (ExtendedDataObject) parentSource.getRootObject();

			Object sourceInstance = getInputFromParent();
			this.input = (sourceInstance == null) ? null : sourceRootDO.getObjectCreator().getExistingDataObject(sourceInstance);

			if(sourceInstance != null && input == null) {
				// This is probably not a containment data object so explicitly create one
				this.input = sourceRootDO.getObjectCreator().createDataObject(sourceInstance, property.getType(), parentSource, property);				
			}
		} else
			this.input = getInputFromParent();

		validate();
	}

	protected void validate() {
		if(!isDataType()) {
			if(this.input != null && !ExtendedDataObject.class.isAssignableFrom(this.input.getClass()))
				throw new IllegalStateException("Source should be a data object but is not treated as one");
			if(this.output != null && !ExtendedDataObject.class.isAssignableFrom(this.output.getClass()))
				throw new IllegalStateException("Target should be a data object but is not treated as one");			
		} else {
			if(this.input != null && ExtendedDataObject.class.isAssignableFrom(this.input.getClass()))
				throw new IllegalStateException("Source should NOT be a data object but is treated as one");
			if(this.output != null && ExtendedDataObject.class.isAssignableFrom(this.output.getClass()))
				throw new IllegalStateException("Target should NOT be a data object but is treated as one");				
		}		
	}

	public CallInfo(Object from, Object to, CallInfo parent, ExtendedProperty property) {

		this.input = from;
		this.output = to;
		this.parent = parent;
		this.inputProperty = property;

		if(parent != null) {
			operation = parent.getOperation();
			settings = parent.getSettings();
			stage = parent.getStage();
		}
		if(parent != null) {
			this.outputObjectCreator = parent.getOutputObjectCreator();
			this.inputObjectCreator = parent.getInputObjectCreator();
		} else
			this.inputObjectCreator = ((ExtendedDataObject)from).getObjectCreator();

		validate();		
	}

	public boolean isCascadable() {
		return AbstractProperty.isCascadable(inputProperty);
	}

	public boolean isRequired() {
		return AbstractProperty.isRequired(getInputProperty());	
	}

	public boolean isCollection() {
		return AbstractProperty.isCollection(getInputProperty());
	}

	public boolean isDataType() {
		if(getInputProperty() == null) {
			// This can be null for a collection element, so check the parent
			if(getParent() != null && getParent().getInputProperty() != null && getParent().getInputProperty().getElementType() != null)
				return getParent().getInputProperty().getElementType().isDataType();
			
			if(input != null)
				return !ExtendedDataObject.class.isAssignableFrom(input.getClass());
			if(output != null)
				return !ExtendedDataObject.class.isAssignableFrom(output.getClass());

		} else 
			return getInputProperty().isDataType();

		return true;
	}

	public String getInputPropertyPath() {
		
		// TODO: Needs optimization
		String result = (inputProperty != null) ? inputProperty.getName() : "";
		CallInfo parentInfo = this.parent;
		while(parentInfo != null) {
			result = ((parentInfo.getInputProperty() != null) ? (parentInfo.getInputProperty().getName() + Settings.PATH_DELIMITER) : "") + result;
			parentInfo = parentInfo.parent;
		}

		return result;
		
		
		/*
		 * We cannot look at the object path since the same object might be referenced from multiple places
		 *
		ExtendedDataObject targetDO = (ExtendedDataObject) getTarget();
		return targetDO.getObjectPath();
		*/
	}

	public boolean isValid() {
		// TODO: check if the source property is valid to be processed
		// This involves checking the scope, detail and view parameters
		// Also the Cascade level

		return true;
	}

	public Operation getOperation() {
		return operation;
	}

	public void setOperation(Operation operation) {
		this.operation = operation;

		if(outputObjectCreator == null || inputObjectCreator == null)
			throw new IllegalStateException("The source and target object creators need to be set in order to create the root target object");

		if(output != null)
			throw new IllegalStateException("The root target data object needs to be created before using the operation object");
	}

	public void linkOutputToParent(Object targetValue) {
		ExtendedType targetType = (ExtendedType) ((ExtendedDataObject) getParent().getOutput()).getType();
		ExtendedProperty targetProperty = (ExtendedProperty) targetType.getProperty(getInputProperty().getName());
		targetProperty.setValue(getParent().getOutput(), targetValue);			
	}

	public Object getOutputFromParent() {
		ExtendedType targetType = (ExtendedType) ((ExtendedDataObject) getParent().getOutput()).getType();
		ExtendedProperty targetProperty = (ExtendedProperty) targetType.getProperty(getInputProperty().getName());
		if(targetProperty == null)
			throw new RuntimeException("Property " + getInputProperty().getName() + " is missing from the derived type.");
		
		return targetProperty.getValue(getParent().getOutput());			
	}

	public ExtendedProperty getOutputProperty() {
		if(getInputProperty() == null)
			return null;
		
		Type targetType = ((ExtendedDataObject) getParent().getOutput()).getType();
		ExtendedProperty targetProperty = (ExtendedProperty) targetType.getProperty(getInputProperty().getName());

		return targetProperty;
	}

	public Object getInputFromParent() {
		return ((ExtendedProperty)getInputProperty()).getValue(getParent().getInput());			
	}

	public ExtendedDataObject getOutputRoot() {
		if(getParent() == null) {
			if(getOutput() == null)
				throw new IllegalStateException("Target root is null, it needs to be created first");

			if(!ExtendedDataObject.class.isAssignableFrom(getOutput().getClass()))
				throw new IllegalStateException("Target root needs to be a data object");

			if ( !((ExtendedDataObject)getOutput()).isRoot() )
				throw new IllegalStateException("Target with null parent needs to be the root data object");

			return (ExtendedDataObject)getOutput();
		} else
			return getParent().getOutputRoot();
	}
	
	public ExtendedDataObject getInputRoot() {
		if(getParent() == null) {
			if(getInput() == null)
				throw new IllegalStateException("Source root is null, set the source before calling this method.");

			if(!ExtendedDataObject.class.isAssignableFrom(getInput().getClass()))
				throw new IllegalStateException("Source root needs to be a data object");

			if ( !((ExtendedDataObject)getInput()).isRoot() )
				throw new IllegalStateException("Source with null parent needs to be the root data object");

			return (ExtendedDataObject)getInput();
		} else
			return getParent().getInputRoot();
		
	}

	public void setOutputObjectCreator(ObjectCreator oc) {
		this.outputObjectCreator = oc;
	}

	public ObjectCreator getOutputObjectCreator() {
		return this.outputObjectCreator;
	}

	public ProcessingStage getStage() {
		return stage;
	}

	public CallInfo setStage(ProcessingStage stage) {
		this.stage = stage;
		return this;
	}

	public boolean isPathSuffix(String pathSuffix) {
		// Walk through the parent chain to see that the path meets the pathSuffix
		if(getInputPropertyPath().endsWith(pathSuffix))
			return true;
		else
			return false;
	}

	public boolean isExternal() {
		boolean result = !this.isCollection() && !this.isCascadable();
		for(AssociationSetting setting: settings.getAssociationSettings()) {
			if(setting.isAggregatePart(this)) {
				result = false;
				break;
			}
		}
		return result;
	}

	public ObjectCreator getInputObjectCreator() {
		return inputObjectCreator;
	}

	public void setInputObjectCreator(ObjectCreator sourceObjectCreator) {
		this.inputObjectCreator = sourceObjectCreator;
	}

	public void clearVisitedOutputs() {
		getOutputObjectCreator().clearVisited();
	}
}
