/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model;

import java.util.ArrayList;
import java.util.List;

import javax.persistence.metamodel.Attribute.PersistentAttributeType;
import javax.xml.bind.annotation.XmlTransient;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import ws.ladder.service.DataAccessService;
import ws.ladder.util.ClassUtil;

public class DerivedProperty extends AbstractProperty {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());	

	private ExtendedProperty referenceProperty;
	
	public DerivedProperty(ExtendedProperty referenceProperty, Type type, DerivedType parentType) {
		super(type, parentType);
		
		this.referenceProperty = referenceProperty;
		
		init();
	}
	
	public ExtendedProperty getReferenceProperty() {
		return referenceProperty;
	}
	
	public Class<?> getJavaType() {
		return (field != null) ? field.getDeclaringClass() : getterMethod.getDeclaringClass();
	}
	
	@Override
	public String getName() {
		return referenceProperty.getName();
	}

	@Override
	public boolean isMany() {
		return referenceProperty.isMany();
	}

	@Override
	public boolean isContainment() {
		return referenceProperty.isContainment();
	}

	@Override
	public Object getDefault() {
		return null;
	}

	@Override
	public boolean isReadOnly() {
		// Don't make assumptions on derived type as this is a separate type
		return false;
	}

	@Override
	public boolean isNullable() {
		return false;
	}

	@Override
	public boolean isOpenContent() {
		return false;
	}

	@Override
	public List<?> getInstanceProperties() {
		return new ArrayList<Object>();
	}

	@Override
	public Object get(Property property) {
		return null;
	}
	
	@Override
	public Property getPositionProperty() {
		return getPositionProperty();
	}	

	@Override
	public void init(DataAccessService das) {
		if(referenceProperty.getKeyType() != null)
			keyType = das.getDerivedType(das.getTypeMapper().toDerived(referenceProperty.getKeyType().getInstanceClass()));
		if(referenceProperty.getElementType() != null)
			elementType = das.getDerivedType(das.getTypeMapper().toDerived(referenceProperty.getElementType().getInstanceClass()));
		
		if( (field != null && ClassUtil.isWrapperType(field.getType())) || 
				(getterMethod != null && ClassUtil.isWrapperType(getterMethod.getReturnType()))
			)
			logger.info("Primitive type found: " + getContainingType().getInstanceClass().getName() + "::" + field.getName() + ", use a wrapper class instead.");
		
		if(referenceProperty.getPositionProperty() != null)
			positionProperty = (ExtendedProperty) getElementType().getProperty(referenceProperty.getPositionProperty().getName());		
	}
	
	@Override
	public boolean needsInitialization() {
		return referenceProperty.needsInitialization();
	}

	@Override
	public void initMappedBy(DataAccessService das) {
		String mappedBy = ClassUtil.getMappedBy(referenceProperty.getContainingType().getInstanceClass(), getName());

		Type type = isMany() ? getElementType() : getType();
		setMappedBy( mappedBy != null ? type.getProperty(mappedBy) : null, mappedBy);

		if(getMappedBy() != null)
			logger.debug("Opposite of property '" + referenceProperty.getContainingType().getName() + "." + getName() + "' is '" + mappedBy + "'");
	}

	@Override
	public PersistentAttributeType getAssociationType() {
		return referenceProperty.getAssociationType();
	}

	@Override
	public boolean isTransient() {
		// check if the annotation is on the getter method
		if(getterMethod != null && getterMethod.getAnnotation(XmlTransient.class) != null)
			return true;	
		
		// check if the annotation is on the field			
		if(field != null && field.getAnnotation(XmlTransient.class) != null)
			return true;
		
		return false;	
	}

	@Override
	public boolean isMap() {
		return referenceProperty.isMap();
	}

	@Override
	public boolean isList() {
		return referenceProperty.isList();
	}

	@Override
	public boolean isSet() {
		return referenceProperty.isSet();
	}	
}
