/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model;

import java.util.List;
import java.util.Set;

import javax.persistence.AccessType;

public interface ExtendedType extends BasicType {
	
	/**
	 * Should the access be based on property (getter method) or 
	 * by field.
	 * 
	 * @return
	 */
	public AccessType getAccessType();
	
	/**
	 * This is used to represent the Java class with meta data (e.g., JPA annotations) and business logic.
	 * 
	 * Types that are based off reference types are called derived types. They are based off a different Java class,
	 * but share the same properties and meta data with the reference type.
	 * @return true if reference type, false if derived type
	 */
	public boolean isReferenceType();
	
	/**
	 * For derived types, this will return the reference type from which it is based off.
	 * For reference types, this will return self.
	 * @return
	 */
	public ExtendedType getReferenceType();
	
	/**
	 * Returns the property that refers to the id of the entity
	 * @return
	 */
	public Property getIdentifierProperty();
	
	/**
	 * Returns the property that is marked as the user key
	 * @return
	 */
	public Property getUserKey();	
	
	/**
	 * Returns the property that is marked as the collection user key
	 * @return
	 */
	public Property getCollectionUserKey();	
	
	/**
	 * Returns the property that refers to the version of the entity
	 * @return
	 */
	public Property getVersionProperty();	
	
	/**
	 * Used to get access to the Reference type from derived types
	 * @return
	 */
	public TypeMapper getTypeMapper();	

	/**
	 * Get a list of all the types that are embedded in this type
	 * @return
	 */
	public List<Type> getEmbeddableTypes();
	
	/**
	 * Return true if the contents of this object cannot be changed
	 */
	public boolean isImmutable();
	
	/**
	 * Returns true if this type is an embedded type
	 */
	public boolean isEmbedded();
	
	/**
	 * Returns the highest ancestor entity in an inheritance hierarchy. Useful in EntityKey
	 */
	public ExtendedType getRootEntityType();
	
	/**
	 * Executes all the PostLogic annotated methods associated with objects of this type and its ancestor types
	 */
	public void invokePostLogic(Object object);

	/**
	 * Returns the name that the persistence provider uses to map an entity
	 * @see getName()
	 * @return
	 */
	public String getEntityName();

	/** 
	 * Returns true if this type represents an entity
	 * @return
	 */
	public boolean isEntity();

	/**
	 * Find all the types of the subclasses of the instance class of this type 
	 */
	public void defineSubtypes(List<ExtendedType> entityTypes);

	/**
	 * Get a list of all the sub types of this type
	 * @return
	 */
	public Set<ExtendedType> getSubtypes();

	/**
	 * Retrieve a property by its alias name
	 * @param name
	 * @return
	 */
	public Property getPropertyByAlias(String name);

	
	/** 
	 * Initialize the position property for list and map types
	 */
	public void initPositionProperty();
		
}
