/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model;

import java.lang.annotation.Annotation;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.persistence.Basic;
import javax.persistence.CascadeType;
import javax.persistence.CollectionTable;
import javax.persistence.Column;
import javax.persistence.ElementCollection;
import javax.persistence.Embedded;
import javax.persistence.EmbeddedId;
import javax.persistence.Enumerated;
import javax.persistence.GeneratedValue;
import javax.persistence.Id;
import javax.persistence.JoinColumn;
import javax.persistence.JoinColumns;
import javax.persistence.JoinTable;
import javax.persistence.Lob;
import javax.persistence.ManyToMany;
import javax.persistence.ManyToOne;
import javax.persistence.MapKey;
import javax.persistence.OneToMany;
import javax.persistence.OneToOne;
import javax.persistence.OrderBy;
import javax.persistence.OrderColumn;
import javax.persistence.PostLoad;
import javax.persistence.PostPersist;
import javax.persistence.PostRemove;
import javax.persistence.PostUpdate;
import javax.persistence.PrePersist;
import javax.persistence.PreRemove;
import javax.persistence.PreUpdate;
import javax.persistence.SequenceGenerator;
import javax.persistence.TableGenerator;
import javax.persistence.Temporal;
import javax.persistence.Transient;
import javax.persistence.Version;
import javax.persistence.metamodel.Attribute;
import javax.persistence.metamodel.Attribute.PersistentAttributeType;
import javax.persistence.metamodel.MapAttribute;
import javax.persistence.metamodel.PluralAttribute;
import javax.persistence.metamodel.PluralAttribute.CollectionType;
import javax.persistence.metamodel.SingularAttribute;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import ws.ladder.service.DataAccessService;
import ws.ladder.service.JPADataAccessService;
import ws.ladder.util.ClassUtil;

public class JPAProperty extends AbstractProperty {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());

	private Attribute<?, ?> attribute;
	private boolean         readOnly;
	private boolean         isPropertyMapped; // Are the JPA mappings on the getter	
	private boolean         isFieldMapped;    // Are the JPA mappings on the field
	private boolean         hasIdAnnotation;
	private boolean         hasVersionAnnotation;	
	private boolean         cascaded;

	public JPAProperty(Attribute<?, ?> attribute, Type type, JPAType parentType) {
		super(type, parentType);
		this.attribute = attribute;

		init();
	}	

	public boolean isPropertyMapped() {
		return isPropertyMapped;
	}	

	public boolean isFieldMapped() {
		return isFieldMapped;
	}

	public boolean isIdentifier() {
		return hasIdAnnotation;
	}

	public boolean isVersion() {
		return hasVersionAnnotation;
	}	

	protected List<Class<? extends Annotation>> getJPAAnnotationList() {
		List<Class<? extends Annotation>> result = new ArrayList();

		// Listing common annotations on top
		result.add(Id.class);		
		result.add(OneToOne.class);
		result.add(OneToMany.class);
		result.add(ManyToMany.class);
		result.add(ManyToOne.class);
		result.add(Version.class);
		result.add(Basic.class);		
		result.add(Column.class);		
		result.add(SequenceGenerator.class);
		result.add(TableGenerator.class);
		result.add(Embedded.class);
		result.add(EmbeddedId.class);
		result.add(Transient.class);
		result.add(ElementCollection.class);
		result.add(GeneratedValue.class);
		result.add(MapKey.class);
		result.add(OrderBy.class);
		result.add(OrderColumn.class);
		result.add(PrePersist.class);
		result.add(PostPersist.class);
		result.add(PreRemove.class);
		result.add(PostRemove.class);
		result.add(PreUpdate.class);
		result.add(PostUpdate.class);
		result.add(PostLoad.class);
		result.add(JoinTable.class);
		result.add(CollectionTable.class);
		result.add(Lob.class);
		result.add(Temporal.class);
		result.add(Enumerated.class);
		result.add(JoinColumn.class);
		result.add(JoinColumns.class);

		return result;
	}

	@Override
	protected void init() {
		super.init();

		// For more details refer to JPA 2.0 specification at section 2.3
		for(Class<? extends Annotation> annotationClass: getJPAAnnotationList()) {
			if(getterMethod.isAnnotationPresent(annotationClass)) {
				isPropertyMapped = true;
				break;
			}
		}	

		for(Class<? extends Annotation> annotationClass: getJPAAnnotationList()) {
			if(field != null && field.isAnnotationPresent(annotationClass)) {
				isFieldMapped = true;
				break;
			}
		}		

		if(isPropertyMapped && isFieldMapped)
			throw new IllegalStateException("Both field and getter method have JPA mappings!");

		// Check if it is an Identifier property
		hasIdAnnotation = isAnnotationPresent(Id.class);

		// Check if it is a version property
		hasVersionAnnotation = isAnnotationPresent(Version.class);

		initContainment();
	}

	protected boolean isAnnotationPresent(Class annotationClass) {
		if(isPropertyMapped && getterMethod.isAnnotationPresent(annotationClass))
			return true;
		else if(isFieldMapped && field != null && field.isAnnotationPresent(annotationClass))
			return true;

		return false;
	}

	@Override	
	protected Annotation getAnnotation(Class annotationClass) {
		if(isPropertyMapped && getterMethod.isAnnotationPresent(annotationClass))
			return getterMethod.getAnnotation(annotationClass);
		else if(isFieldMapped && field != null && field.isAnnotationPresent(annotationClass))
			return field.getAnnotation(annotationClass);

		return null;
	}	

	protected boolean isCascaded(CascadeType[] cascadeTypes) {
		List<CascadeType> types = Arrays.asList(cascadeTypes);
		return types.contains(CascadeType.ALL) || types.contains(CascadeType.PERSIST) || types.contains(CascadeType.MERGE);
	}

	protected void initContainment() {
		// The attribute.isAssociation method is not working, so we have to do this

		if(attribute.getPersistentAttributeType() == PersistentAttributeType.ONE_TO_ONE) {
			if(isPropertyMapped && getterMethod.isAnnotationPresent(OneToOne.class)) {
				OneToOne oneToOne = getterMethod.getAnnotation(OneToOne.class);
				cascaded = isCascaded(oneToOne.cascade());
			} else if(isFieldMapped && field != null && field.isAnnotationPresent(OneToOne.class)) {
				OneToOne oneToOne = getterMethod.getAnnotation(OneToOne.class);
				cascaded = isCascaded(oneToOne.cascade());				
			}
		} else if (attribute.getPersistentAttributeType() == PersistentAttributeType.ONE_TO_MANY) {
			if(isPropertyMapped && getterMethod.isAnnotationPresent(OneToMany.class)) {
				OneToMany oneToMany = getterMethod.getAnnotation(OneToMany.class);
				cascaded = isCascaded(oneToMany.cascade());
			} else if(isFieldMapped && field != null && field.isAnnotationPresent(OneToMany.class)) {
				OneToMany oneToMany = getterMethod.getAnnotation(OneToMany.class);
				cascaded = isCascaded(oneToMany.cascade());				
			}			
		} else if (attribute.getPersistentAttributeType() == PersistentAttributeType.MANY_TO_ONE) {
			if(isPropertyMapped && getterMethod.isAnnotationPresent(ManyToOne.class)) {
				ManyToOne manyToOne = getterMethod.getAnnotation(ManyToOne.class);
				cascaded = isCascaded(manyToOne.cascade());
			} else if(isFieldMapped && field != null && field.isAnnotationPresent(ManyToOne.class)) {
				ManyToOne manyToOne = getterMethod.getAnnotation(ManyToOne.class);
				cascaded = isCascaded(manyToOne.cascade());				
			}			
		} else if (attribute.getPersistentAttributeType() == PersistentAttributeType.MANY_TO_MANY) {
			if(isPropertyMapped && getterMethod.isAnnotationPresent(ManyToMany.class)) {
				ManyToMany manyToMany = getterMethod.getAnnotation(ManyToMany.class);
				cascaded = isCascaded(manyToMany.cascade());
			} else if(isFieldMapped && field != null && field.isAnnotationPresent(ManyToMany.class)) {
				ManyToMany manyToMany = getterMethod.getAnnotation(ManyToMany.class);
				cascaded = isCascaded(manyToMany.cascade());				
			}			
		}
	}

	@Override
	public void init(DataAccessService das) {
		if( PluralAttribute.class.isAssignableFrom(attribute.getClass()) ) {
			PluralAttribute<?, ?, ?> pluralAttribute = (PluralAttribute<?, ?, ?>) attribute;
			elementType = das.getType(pluralAttribute.getElementType().getJavaType());
			if(MapAttribute.class.isAssignableFrom(attribute.getClass())) {
				MapAttribute<?, ?, ?> mapAttribute = (MapAttribute<?, ?, ?>) attribute;
				keyType = das.getType(mapAttribute.getKeyJavaType());
			}
		}		
	}	

	@Override
	public String getName() {
		return attribute.getName();
	}

	@Override
	public boolean isMany() {
		return attribute.isCollection();
	}

	@Override
	public boolean isContainment() {
		if(cascaded)
			return true;

		if(isDataType()) {
			if(isMany()) {
				if((ExtendedType.class.isAssignableFrom(getElementType().getClass())))
					return ((ExtendedType)getElementType()).isEmbedded();
			}
			return true;
		} else {
			if((ExtendedType.class.isAssignableFrom(getType().getClass())))
				return ((ExtendedType)getType()).isEmbedded();
		}

		return false;
	}

	/**
	 * Not directly supported in Hibernate
	 */
	@Override
	public Object getDefault() {
		return null;
	}

	@Override
	public boolean isReadOnly() {
		return readOnly;
	}

	public void setReadOnly(JPADataAccessService das) {
		readOnly = !ClassUtil.isUpdatable(getContainingType().getInstanceClass(), getName());	
	}

	@Override
	public boolean isNullable() {
		if(SingularAttribute.class.isAssignableFrom(attribute.getClass()))
			return ((SingularAttribute<?, ?>)attribute).isOptional();
		else
			return true;
	}

	@Override
	public boolean isOpenContent() {
		return false;
	}

	@Override
	public List<?> getInstanceProperties() {
		return new ArrayList<Object>();
	}

	@Override
	public Object get(Property property) {
		return null;
	}

	@Override
	public void initMappedBy(DataAccessService das) {
		String mappedBy = ClassUtil.getMappedBy(getContainingType().getInstanceClass(), getName());

		Type type = isMany() ? getElementType() : getType();
		setMappedBy( mappedBy != null ? type.getProperty(mappedBy) : null, mappedBy);

		if(getMappedBy() != null)
			logger.debug("Opposite of property '" + getContainingType().getName() + "." + getName() + "' is '" + mappedBy + "'");
	}

	@Override
	public PersistentAttributeType getAssociationType() {
		return attribute.getPersistentAttributeType();
	}

	@Override
	public boolean isMap() {
		if(isMany()) {
			if( PluralAttribute.class.isAssignableFrom(attribute.getClass()) ) {
				PluralAttribute<?, ?, ?> pluralAttribute = (PluralAttribute<?, ?, ?>) attribute;
				if(pluralAttribute.getCollectionType() == CollectionType.MAP )
					return true;
			}
		}

		return false;
	}

	@Override
	public boolean isList() {
		if(isMany() && isAnnotationPresent(OrderColumn.class)) {
			if( PluralAttribute.class.isAssignableFrom(attribute.getClass()) ) {
				PluralAttribute<?, ?, ?> pluralAttribute = (PluralAttribute<?, ?, ?>) attribute;
				if(pluralAttribute.getCollectionType() == CollectionType.LIST )
					return true;
			}
		}

		return false;
	}

	@Override
	public boolean isSet() {
		if(isMany()) {
			if( PluralAttribute.class.isAssignableFrom(attribute.getClass()) ) {
				PluralAttribute<?, ?, ?> pluralAttribute = (PluralAttribute<?, ?, ?>) attribute;
				if(pluralAttribute.getCollectionType() == CollectionType.SET )
					return true;
			}
		}

		return false;
	}		
}
