/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import javax.persistence.AccessType;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.hibernate.metamodel.binding.AttributeBinding;
import org.hibernate.metamodel.binding.BasicAttributeBinding;
import org.hibernate.metamodel.binding.EntityBinding;
import org.hibernate.metamodel.binding.EntityIdentifier;
import org.hibernate.metamodel.domain.Attribute;

import ws.ladder.service.MetadataDAS;
import ws.ladder.util.ClassUtil;

public class MetadataType extends AbstractType {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());	

	private org.hibernate.type.Type        hibernateType;
	private EntityBinding                  entityBinding;
	private List<Type>                     baseTypes;
	private MetadataProperty               identifierProperty;
	private MetadataProperty               versionProperty;
	private AccessType                     accessType;

	public MetadataType(org.hibernate.type.Type hibernateType, EntityBinding entityBinding) {
		super();

		this.hibernateType = hibernateType;
		this.entityBinding = entityBinding;
		accessType = ClassUtil.getHibernateAccessType(getInstanceClass());

		initImmutable();
	}

	@Override
	public String getName() {
		// The name is unique in the namespace because if is qualified by the package name
		return hibernateType.getName();
	}
	
	@Override
	public String getEntityName() {
		return hibernateType.getName();
	}

	@Override
	public String getURI() {
		return null;
	}

	@Override
	public Class<?> getInstanceClass() {
		return hibernateType.getReturnedClass();
	}

	@Override
	public boolean isInstance(Object object) {
		return getClass().isAssignableFrom(object.getClass());
	}

	/**
	 * The property names within the type are unique
	 */
	@Override
	public List<Property> getProperties() {
		return sort(new ArrayList<Property>(properties.values()));
	}

	public void setProperty(MetadataDAS dataAccessService) {
		if(properties == null) {
			// populate the properties for this type
			properties = new HashMap<String, Property>();	

			//Entity entity = entityBinding.getEntity();
			//Iterator<Attribute> propertyIterator = entity.attributes().iterator();
			Iterator<AttributeBinding> propertyIterator = entityBinding.getAttributeBindingClosure().iterator();
			while(propertyIterator.hasNext()) {
				Attribute attribute = (propertyIterator.next()).getAttribute();

				logger.debug("[" + getName() + "] Metadata Property name: " + attribute.getName() + ", type name: " + attribute.getAttributeContainer().getClassName());
				Type propertyType = dataAccessService.getType(MetadataProperty.getClass(attribute));

				MetadataProperty property = new MetadataProperty(attribute, propertyType, this);
				properties.put(property.getName(), property);
				property.init(dataAccessService);
			}	

			EntityIdentifier entityIdentifier = entityBinding.getHierarchyDetails().getEntityIdentifier();
			if(entityIdentifier != null) {
				logger.debug("Metadata Identifier attribute name: " + entityIdentifier.getValueBinding().getAttribute().getName());				
				String IdentifierName = entityIdentifier.getValueBinding().getAttribute().getName();
				identifierProperty = (MetadataProperty) properties.get(IdentifierName);
			}
			
			BasicAttributeBinding versionAttributeBinding = entityBinding.getHierarchyDetails().getVersioningAttributeBinding();		
			if(versionAttributeBinding != null) {
				logger.debug("Metadata version attribute name: " + versionAttributeBinding.getAttribute().getName());				
				String versionName = versionAttributeBinding.getAttribute().getName();
				versionProperty = (MetadataProperty) properties.get(versionName);
			}					
		} 		
	}

	public void setOpposite(MetadataDAS dataAccessService) {
		for(Property property: properties.values()) {
			((MetadataProperty)property).initMappedBy(dataAccessService);
		}
	}

	@Override
	public Property getProperty(String propertyName) {
		return properties.get(propertyName);
	}

	@Override
	public boolean isOpen() {
		return false;
	}

	@Override
	public boolean isSequenced() {
		return false;
	}

	@Override
	public boolean isAbstract() {
		return false;
	}

	@Override
	public List<Type> getBaseTypes() {
		return baseTypes;
	}

	public void setBaseType(List<Type> types) {
		baseTypes = types;
	}

	@Override
	public List<Property> getDeclaredProperties() {
		// TODO: check if the Entity#attributes method returns the declared list
		//System.out.println("[" + getName() + "] declared property name: " + declaredProperty.getName());
		return new ArrayList<Property>();
	}

	@Override
	public List<?> getAliasNames() {
		return new ArrayList<String>();
	}

	@Override
	public List<?> getInstanceProperties() {
		return new ArrayList<Object>();
	}

	@Override
	public Object get(Property property) {
		return null;
	}

	public EntityBinding getEntityBinding() {
		return entityBinding;
	}

	@Override
	public AccessType getAccessType() {
		return accessType;
	}

	@Override
	public Property getIdentifierProperty() {
		return identifierProperty;
	}

	@Override
	public boolean isEmbedded() {
		return hibernateType.isComponentType();
	}
	
	@Override
	public boolean isEntity() {
		return hibernateType.isEntityType(); 
	}			

	@Override
	public Property getVersionProperty() {
		return versionProperty;
	}	
}
