/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model.action;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import ws.ladder.model.CallInfo;
import ws.ladder.model.ExtendedDataObject;


public abstract class CollectionUpdateAction implements Executable {
	protected Set<AddElementAction> addByOppositeActions = new HashSet<AddElementAction>();
	protected Set<RemoveElementAction> removeByOppositeActions = new HashSet<RemoveElementAction>();	

	public abstract void addAction(Executable action);

	public abstract void addOrReposition(CallInfo callInfo) throws Exception;
	
	public ExtendedDataObject getTriggerObject() {
		throw new UnsupportedOperationException("It is not meaningful to call this method on a CollectionUpdateAction since a collection can be updated from multiple sources and does not represent a single execution");
	}

	public void addTriggeredByOppositeAction(ElementAction action) {
		if(AddElementAction.class.isAssignableFrom(action.getClass()))
			addByOppositeActions.add((AddElementAction) action);
		else if(RemoveElementAction.class.isAssignableFrom(action.getClass()))
			removeByOppositeActions.add((RemoveElementAction) action);
	}

	@Override
	public Executable getTriggeringAction() {
		return null;
	}

	public void removeObsoletes(CallInfo callInfo) {

		/* algorithm
		 * 1. All the objects in the targetCollection should have ids since these are entities
		 * 2. We get a list of all ids from the sourceCollection
		 * 3. All ids that are not present in the sourceCollection are obsolete and needs to be removed
		 * 
		 * For a Map, we only honor entity in the value and not in the key. So we just need to check the values collection
		 * of the source and target maps.
		 */		

		Map targetKeys = new HashMap<Object, Object>();
		for(ExtendedDataObject element: ((ExtendedDataObject)callInfo.getOutput()).getList(callInfo))
			targetKeys.put(element.getIdentifierValue(), element);		

		Map sourceKeys = new HashMap<Object, Object>();		
		if(callInfo.getInput() != null) 
			for(ExtendedDataObject element: ((ExtendedDataObject)callInfo.getInput()).getList(callInfo)) {
				Object idValue = element.getIdentifierValue();
				if(idValue != null)
					sourceKeys.put(idValue, element);
			}			

		for(Object entry : targetKeys.entrySet()) {
			Object targetKey = ((Map.Entry) entry).getKey();
			Object targetElement = ((Map.Entry) entry).getValue();

			if(!sourceKeys.containsKey(targetKey)) { // obsolete element being marked for removal
				CallInfo next = new CallInfo(null, null, callInfo, null);
				next.setOutput(callInfo.getOutputObjectCreator().createTarget(next, targetElement, null));
				removeCollectionElement(next, (ExtendedDataObject) next.getOutput());
			}
		}		
	}

	protected void removeCollectionElement(CallInfo callInfo, ExtendedDataObject collectionElement) {
		ExtendedDataObject collectionOwner = (ExtendedDataObject) callInfo.getParent().getParent().getOutput();
		
		// TODO: List and Map might need to add additional info to the RemoveElementAction. So this method will need to be overridden by the subclasses.
		PropertyKey propertyKey = new PropertyKey((ExtendedDataObject) callInfo.getParent().getOutput(), callInfo.getParent().getOutputProperty());
		Executable originalAction = new RemoveElementAction(propertyKey, collectionElement, collectionOwner, null, null, callInfo);
		this.addAction(originalAction);		

		// Set the backRef value to null. This is an asynchronous activity
		PropertyKey newOppositeKey = new PropertyKey(collectionElement, callInfo.getParent().getOutputProperty().getOpposite());			
		callInfo.getOutputRoot().getDependencyGraph().addAction(new SetterAction(null, newOppositeKey, originalAction)); 	

	}		
}
