/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model.action;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import ws.ladder.model.AggregateAction;
import ws.ladder.model.CallInfo;
import ws.ladder.model.ExtendedDataObject;
import ws.ladder.model.ExtendedProperty;
import ws.ladder.model.event.CollectionElement;
import ws.ladder.model.event.CollectionEvent;
import ws.ladder.model.operation.AbstractOperation;
import ws.ladder.util.ClassUtil;


public class ListUpdateAction extends CollectionUpdateAction {

	private PropertyKey key;
	private List<AddElementAction>        addActions = new ArrayList<AddElementAction>();
	private List<RemoveElementAction>     removeActions = new ArrayList<RemoveElementAction>();
	private List<RepositionElementAction> repositionActions = new ArrayList<RepositionElementAction>(); 

	public ListUpdateAction(PropertyKey key) {
		this.key = key;
	}

	@Override
	public void execute() {
		boolean collectionModified = false;

		// TODO: remove elements

		// Sort by position
		Map<Integer, AddElementAction> listOrderedActions = new TreeMap<Integer, AddElementAction>();
		
		for(AddElementAction action: addActions) 
			listOrderedActions.put(Integer.parseInt(action.getPosition().toString()), action);
		// TODO: Handler conflicts
		for(AddElementAction action: addByOppositeActions) 
			listOrderedActions.put(Integer.parseInt(action.getPosition().toString()), action);		

		for(AddElementAction action: listOrderedActions.values()) {
			// Add the element to the set
			customAddToList(action);		
			collectionModified = true;			
		}

		// checkOutOfSync();

		if(collectionModified)
			AbstractOperation.setModified(key.getDataObject());		
	}

	@Override
	public void addAction(Executable action) {
		if(AddElementAction.class.isAssignableFrom(action.getClass()))
			addActions.add((AddElementAction) action);
		else if(RemoveElementAction.class.isAssignableFrom(action.getClass()))
			removeActions.add((RemoveElementAction) action);
		else if(RepositionElementAction.class.isAssignableFrom(action.getClass()))
			repositionActions.add((RepositionElementAction) action);		
		else
			throw new IllegalArgumentException("A list can have only add, remove and reposition actions");
	}

	@Override
	public PropertyKey getKey() {
		return key;
	}

	public void addOrReposition(CallInfo callInfo) throws Exception {
		Map<Object, ExtendedDataObject> targetKeys = new HashMap<Object, ExtendedDataObject>();
		for(ExtendedDataObject element: ((ExtendedDataObject)callInfo.getOutput()).getList(callInfo))
			targetKeys.put(element.getIdentifierValue(), element);		

		if(callInfo.getInput() != null) {
			AbstractOperation oper = (AbstractOperation) callInfo.getOperation();	
			Iterator<ExtendedDataObject> elements = ((ExtendedDataObject)callInfo.getInput()).getList(callInfo).iterator();
			for(int i = 0; elements.hasNext(); i++) {
				ExtendedDataObject childDO = elements.next();
				CallInfo next = new CallInfo(childDO, null, callInfo, null);

				ExtendedDataObject targetElement = targetKeys.get(childDO.getIdentifierValue());
				if(callInfo.getSettings().getAction() != AggregateAction.CREATE && targetElement != null) { // updating an existing object

					next.setOutput(targetElement);
					oper.processValue(next);

				} else {

					next.setOutput(oper.createTarget(next, null));
					oper.processValue(next);

					// AddElementAction
					addListElement(next, (ExtendedDataObject) next.getOutput(), i);
				}
				// Add edge in dependency graph
				AbstractOperation.addEdge(next);	
			}	
		}
	}

	protected void addListElement(CallInfo callInfo, ExtendedDataObject collectionElement, Object position) {
		Object collection = callInfo.getParent().getOutput();
		ExtendedDataObject collectionOwner = (ExtendedDataObject) callInfo.getParent().getParent().getOutput();

		PropertyKey propertyKey = new PropertyKey((ExtendedDataObject) collection, callInfo.getParent().getOutputProperty());
		Executable originalAction = new AddElementAction(propertyKey, collectionElement, collectionOwner, null, position, callInfo);
		this.addAction(originalAction);		

		// Set the backRef of value to point to the parent target (or to the ancester in case of embedded property)
		// TODO: handle MANY_TO_MANY backRef
		if(callInfo.getParent().getOutputProperty().isBiDirectional()) {
			PropertyKey newOppositeKey = new PropertyKey(collectionElement, callInfo.getParent().getOutputProperty().getOpposite());			
			callInfo.getOutputRoot().getDependencyGraph().addAction(new SetterAction(collectionOwner, newOppositeKey, originalAction));
		}
	}

	protected void customAddToList(AddElementAction action) {
		CallInfo callInfo = action.getCallInfo();

		ExtendedDataObject collectionOwner  = action.getCollectionOwner();
		Object collection                   = ((ExtendedProperty) action.getKey().getProperty()).getValue(collectionOwner);
		Object collectionElement            = action.getCollectionElement();
		ExtendedProperty collectionProperty = (ExtendedProperty) action.getKey().getProperty();

		// It is possible for the collection object to be null, if this is due to the MANY_TO_ONE end only being set
		try {
			if(collection == null) {
				ExtendedDataObject collectionDataObject = (ExtendedDataObject) collectionOwner.createDataObject((Object)null, collectionProperty.getType());
				collection = collectionDataObject.getInstance();
				((ExtendedProperty) action.getKey().getProperty()).setValue(collectionOwner, collection);
			} 
		} catch(Exception e) {
			throw ClassUtil.wrapRun(e);
		}


		// check if there is a custom collection add
		Method customCollectionAdd = ClassUtil.getCustomAdd(collectionProperty.getElementType().getInstanceClass(), collectionProperty.getName());
		if(customCollectionAdd != null) {
			CollectionEvent event = new CollectionElement(callInfo, collection, collectionElement);
			collectionProperty.setCustomAdd(collection, event);
		} else {
			if(((java.util.Collection<Object>)collection).size() == Integer.parseInt(action.getPosition().toString()) )
				((java.util.Collection<Object>)collection).add(ClassUtil.getInstance(collectionElement));
			else
				throw new IllegalArgumentException("Trying to add a list element at the wrong position: " + action.getPosition() + ", list size: " + 
						((java.util.Collection<Object>)collection).size());
		}
	}		
}
