/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model.action;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import ws.ladder.exception.BidirOutOfSyncException;
import ws.ladder.model.AggregateAction;
import ws.ladder.model.CallInfo;
import ws.ladder.model.ExtendedDataObject;
import ws.ladder.model.ExtendedProperty;
import ws.ladder.model.event.CollectionElement;
import ws.ladder.model.event.CollectionEvent;
import ws.ladder.model.operation.AbstractOperation;
import ws.ladder.util.ClassUtil;


public class SetUpdateAction extends CollectionUpdateAction {

	private PropertyKey key;
	private List<AddElementAction>               addActions = new ArrayList<AddElementAction>();
	private List<RemoveElementAction>            removeActions = new ArrayList<RemoveElementAction>(); 
	private Map<ExtendedDataObject, List<ElementAction>> actionsOnElementMap = new IdentityHashMap<ExtendedDataObject, List<ElementAction>>();

	public SetUpdateAction(PropertyKey key) {
		this.key = key;
	}

	private void mapActionsOnElement(ExtendedDataObject collectionElement, ElementAction action) {
		List<ElementAction> actions = actionsOnElementMap.get(collectionElement);
		if(actions == null) {
			actions = new ArrayList<ElementAction>();
			actionsOnElementMap.put(collectionElement, actions);
		}
		actions.add(action);
	}

	@Override
	public void execute() {
		boolean collectionModified = false;

		// Create a map between the collectionElement and the actions acted upon it

		Set<RemoveElementAction> rActions = new HashSet<RemoveElementAction>(removeActions);
		rActions.addAll(removeByOppositeActions);

		// Remove obsolete elements
		for(RemoveElementAction action: rActions) {
			// Remove the element from the set
			customRemoveFromSet(action);
			mapActionsOnElement(action.getCollectionElement(), action);
			collectionModified = true;
		}

		// Add new elements
		Set<AddElementAction> actions = new HashSet<AddElementAction>(addActions);
		actions.addAll(addByOppositeActions);

		for(AddElementAction action: actions) {
			// Add the element to the set
			customAddToSet(action);
			mapActionsOnElement(action.getCollectionElement(), action);			
			collectionModified = true;			
		}

		checkOutOfSync();

		if(collectionModified)
			AbstractOperation.setModified(key.getDataObject());		
	}

	private void checkOutOfSync() {
		// It is an error to only have a triggered by action doing a removal while the collectionOwner is present in the input and is processed
		for(Map.Entry<ExtendedDataObject, List<ElementAction>> entry: actionsOnElementMap.entrySet()) {
			ElementAction lastAction = null;
			boolean isUnlinkedOnOneSide = false;
			for(ElementAction action: entry.getValue()) {
				if(RemoveElementAction.class.isAssignableFrom(action.getClass()))
					isUnlinkedOnOneSide = true;
				lastAction = action;
				if(action.getTriggeringAction() == null) {
					lastAction = null; // The same action is also present in the owner, so the change is in sync
					break;
				}
			}

			if(lastAction != null && isUnlinkedOnOneSide) { // potential out-of-sync condition
				ExtendedDataObject collectionOwner = lastAction.getCollectionOwner();
				ExtendedDataObject sourceRoot = lastAction.getCallInfo().getInputRoot();

				// is out-of-sync
				ExtendedDataObject sourceCollectionOwner = sourceRoot.getEntity(collectionOwner);
				if(sourceCollectionOwner != null && sourceCollectionOwner.isVisited()) // The owner is present in the input and was processed
					throw new BidirOutOfSyncException(collectionOwner, lastAction.getKey().getProperty(), lastAction.getCollectionElement(), lastAction.getKey().getProperty().getOpposite());
			}
		}		
	}

	protected void customAddToSet(AddElementAction action) {
		CallInfo callInfo = action.getCallInfo();

		ExtendedDataObject collectionOwner  = action.getCollectionOwner();
		Object collection                   = ((ExtendedProperty) action.getKey().getProperty()).getValue(collectionOwner);
		Object collectionElement            = action.getCollectionElement();
		ExtendedProperty collectionProperty = (ExtendedProperty) action.getKey().getProperty();

		// It is possible for the collection object to be null, if this is due to the MANY_TO_ONE end only being set
		try {
			if(collection == null) {
				ExtendedDataObject collectionDataObject = (ExtendedDataObject) collectionOwner.createDataObject((Object)null, collectionProperty.getType());
				collection = collectionDataObject.getInstance();
				((ExtendedProperty) action.getKey().getProperty()).setValue(collectionOwner, collection);
			} 
		} catch(Exception e) {
			throw ClassUtil.wrapRun(e);
		}


		// check if there is a custom collection add
		Method customCollectionAdd = ClassUtil.getCustomAdd(collectionProperty.getElementType().getInstanceClass(), collectionProperty.getName());
		if(customCollectionAdd != null) {
			CollectionEvent event = new CollectionElement(callInfo, collection, collectionElement);
			collectionProperty.setCustomAdd(collection, event);
		} else
			((java.util.Collection<Object>)collection).add(ClassUtil.getInstance(collectionElement));		
	}	

	protected void customRemoveFromSet(RemoveElementAction action) {
		CallInfo callInfo = action.getCallInfo();		

		Object collectionOwner              = action.getCollectionOwner();
		Object collection                   = ((ExtendedProperty) action.getKey().getProperty()).getValue(collectionOwner);
		Object collectionElement            = action.getCollectionElement();
		ExtendedProperty collectionProperty = (ExtendedProperty) action.getKey().getProperty();

		// check if there is a custom collection add
		Method customCollectionRemove = ClassUtil.getCustomRemove(collectionProperty.getElementType().getInstanceClass(), collectionProperty.getName());
		if(customCollectionRemove != null) {
			CollectionEvent event = new CollectionElement(callInfo, collection, collectionElement);
			collectionProperty.setCustomRemove(collection, event);
		} else
			((java.util.Collection<Object>)collection).remove(ClassUtil.getInstance(collectionElement));		
	}		

	@Override
	public void addAction(Executable action) {
		if(AddElementAction.class.isAssignableFrom(action.getClass()))
			addActions.add((AddElementAction) action);
		else if(RemoveElementAction.class.isAssignableFrom(action.getClass()))
			removeActions.add((RemoveElementAction) action);
		else
			throw new IllegalArgumentException("A set can have only add and remove actions");
	}

	public void addOrReposition(CallInfo callInfo) throws Exception {
		Map<Object, ExtendedDataObject> targetKeys = new HashMap<Object, ExtendedDataObject>();
		for(ExtendedDataObject element: ((ExtendedDataObject)callInfo.getOutput()).getList(callInfo))
			targetKeys.put(element.getIdentifierValue(), element);		

		if(callInfo.getInput() != null) {
			AbstractOperation oper = (AbstractOperation) callInfo.getOperation();	
			for(ExtendedDataObject childDO: ((ExtendedDataObject)callInfo.getInput()).getList(callInfo)) {
				CallInfo next = new CallInfo(childDO, null, callInfo, null);

				ExtendedDataObject targetElement = targetKeys.get(childDO.getIdentifierValue());
				if(callInfo.getSettings().getAction() != AggregateAction.CREATE && targetElement != null) { // updating an existing object

					next.setOutput(targetElement);
					oper.processValue(next);

				} else {

					next.setOutput(oper.createTarget(next, null));
					oper.processValue(next);

					// AddElementAction
					addSetElement(next, (ExtendedDataObject) next.getOutput());
				}
				// Add edge in dependency graph
				AbstractOperation.addEdge(next);	
			}	
		}
	}


	protected void addSetElement(CallInfo callInfo, ExtendedDataObject collectionElement) {
		Object collection = callInfo.getParent().getOutput();
		ExtendedDataObject collectionOwner = (ExtendedDataObject) callInfo.getParent().getParent().getOutput();

		PropertyKey propertyKey = new PropertyKey((ExtendedDataObject) collection, callInfo.getParent().getOutputProperty());
		Executable originalAction = new AddElementAction(propertyKey, collectionElement, collectionOwner, null, null, callInfo);
		this.addAction(originalAction);		

		// Set the backRef of value to point to the parent target (or to the ancester in case of embedded property)
		// TODO: handle MANY_TO_MANY backRef
		if(callInfo.getParent().getOutputProperty().isBiDirectional()) {
			PropertyKey newOppositeKey = new PropertyKey(collectionElement, callInfo.getParent().getOutputProperty().getOpposite());			
			callInfo.getOutputRoot().getDependencyGraph().addAction(new SetterAction(collectionOwner, newOppositeKey, originalAction));
		}
	}		


	@Override
	public PropertyKey getKey() {
		return key;
	}

}
