/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model.operation;

import java.lang.reflect.Method;
import java.util.Map;

import ws.ladder.exception.PropertyMappingException;
import ws.ladder.model.AggregateAction;
import ws.ladder.model.CallInfo;
import ws.ladder.model.ContentDetail;
import ws.ladder.model.ExtendedDataObject;
import ws.ladder.model.ExtendedProperty;
import ws.ladder.model.ExtendedType;
import ws.ladder.model.Node;
import ws.ladder.model.ProcessingStage;
import ws.ladder.model.Property;
import ws.ladder.model.event.CollectionElement;
import ws.ladder.model.event.CollectionEvent;
import ws.ladder.util.ClassUtil;

public abstract class AbstractOperation implements Operation {
	
	protected void setResult(Object target) {}
	
	public void execute(CallInfo callInfo) {
		// Create new objects
		this.process(callInfo.setStage(ProcessingStage.CREATE)); 
		callInfo.clearVisitedOutputs();		

		// TODO: get list of revisions for this aggregate and iterate this process for each revision
		// The upgrade and downgrade methods will be called for revision processing instead of the process method
		// Step 1: Create the update actions
		setVisited(callInfo, true);
		this.process(callInfo.setStage(ProcessingStage.UPDATE));
		callInfo.clearVisitedOutputs();		
		// Step 2: Execute the actions
		callInfo.getOutputRoot().getDependencyGraph().processActions(callInfo.getSettings());

		// Process post actions 
		this.process(callInfo.setStage(ProcessingStage.POSTLOGIC));
		callInfo.clearVisitedOutputs();	

		setResult(callInfo.getOutput());
	}
	
	private void setVisited(CallInfo callInfo, boolean value) {
		((Node)callInfo.getOutput()).setVisited(true); // marked nodes we processed
		((Node)callInfo.getInput()).setVisited(true); // This flag on the source is used by the conflict resolution logic		
	}

	protected boolean isAllowed(CallInfo callInfo, ExtendedDataObject source, Property sourceProperty) {
		return (callInfo.getStage() == ProcessingStage.CREATE &&  (((ExtendedType)source.getType()).isImmutable() || 
				sourceProperty.isContainment()                  ||
				sourceProperty.getType().isDataType()))         
				||
				callInfo.getStage() != ProcessingStage.CREATE;

	}

	protected void processPostLogic(CallInfo callInfo) {
		if(callInfo.getStage() == ProcessingStage.POSTLOGIC) {
			ExtendedDataObject target = (ExtendedDataObject) callInfo.getOutput();
			if(callInfo.isCascadable())
				target.invokePostLogic();
		}		
	}

	/* Algorithm
	 * =========
	 * 
	 * 1. Process the "from" data object
	 *    a) If it has properties then process each on in turn
	 *    b) If it does not have properties and is a collection/map then process each element in turn
	 *    c) If it does not have properties then process the data type
	 * 2. If it is a valid property, call ObjectCreator#toDataObject on the "from" property and goto STEP 1
	 * 3. Else set as null in the "to" object and process the next item
	 * 
	 */		
	public void process(CallInfo callInfo) {

		try {
			if(!callInfo.isDataType()) {
				ExtendedDataObject source = (ExtendedDataObject) callInfo.getInput();
				for(Property sourceProperty: source.getType().getProperties()) {
					if(isAllowed(callInfo, source, sourceProperty)) {
						CallInfo next = new CallInfo(null, callInfo, (ExtendedProperty) sourceProperty);
						processValue(next); // recurse
					}
				}

				processPostLogic(callInfo);
			}

		} catch (Exception e) {
			throw ClassUtil.wrapRun(e);
		} 
	}

	private void processToMany(CallInfo callInfo) throws Exception {
		if(callInfo.getOutput() == null)
			return;

		processCollection(callInfo);
	}

	public static void setModified(Object dataObject) {
		ExtendedDataObject obj = (ExtendedDataObject) dataObject;
		while( ((ExtendedType)obj.getType()).isEmbedded() )
			obj = (ExtendedDataObject) obj.getContainer();
		((Node)obj).setModified(true);
	}

	public boolean processPrePropertyLogic(CallInfo ci) {
		// TODO: return true if further processing is necessary, override if necessary
		return true;
	}
	
	public void processValue(CallInfo ci) throws Exception {

		if(!processPrePropertyLogic(ci))
			return;
		
		if(!ci.isValid())
			return;

		if(ci.getInput() == null) {
			if ( ci.getOutputFromParent() != null && shouldUnlink(ci)) {
				processNullValue(ci);
			}

			return;
		}		

		// Skip populating the identifier for JPA object copy, as we want this to be null
		// Subclass can override this behavior
		if(isIdentifier(ci)) // handled by processIdentifier method
			return;		

		// If the property to be copied is specified then we perform the copy only for that property
		if(ci.getInputProperty() != null) { // Owner is not a collection

			// Simply copy the value
			if(ci.isDataType()) {
				ci.setOutput(ci.getOutputObjectCreator().createDataType(ci.getInputFromParent()));
				Object oldTarget = ci.getOutputFromParent();
				if( !( oldTarget == ci.getOutput() || (oldTarget != null && oldTarget.equals(ci.getOutput()))) ) {
					setModified(ci.getParent().getOutput());
					ci.linkOutputToParent(ci.getOutput());
				}
				return;
			}

			// if a reference was returned then we need to not process it anymore
			ci.setOutput(ci.getOutputObjectCreator().getExistingDataObject(ci.getInput())); // TODO: check if processed flag is necessary - What about temp objects with reference relationship?

			if(ci.getOutput() != null) {
				setPropertyTarget(ci, ci.getOutput());

				if( ((Node)ci.getOutput()).isVisited() ) {
					addEdge(ci); // Ensure this path is captured as an edge
					return;				
				}
			}
			else
				// Do the copy here if necessary
				ci.setOutput(setExistingOrNewCopy(ci));

		}

		addEdge(ci);

		// We do not handle collection objects here because they need additional processing on the child elements so, this short circuiting cannot be applied to them.
		if(!ci.getSettings().getAssociationStrategy().doProcess(ci))
			return;

		if(!((Node)ci.getOutput()).isVisited()) {
			setVisited(ci, true);
		} else
			return;

		if(ci.getInputProperty() != null && ci.getInputProperty().isMany()) // Process the collection/map
			processToMany(ci);		
		else
			process(ci);
	}	
	
	public static void addEdge(CallInfo ci) {
		if(ci.getSettings().getAction() != AggregateAction.READ && ci.getStage() == ProcessingStage.UPDATE) {
			((Node) ci.getOutput()).addIncomingEdge((Node) ci.getParent().getOutput(), ci.getOutputProperty());				
		}		
	}

	@Override
	public boolean isExternalAssociationLink(CallInfo ci) {
		boolean result = ci.isExternal();
		if(ci.getSettings().getDetail() == ContentDetail.TEMPLATE)
			result = !ci.isCascadable(); // TEMPLATE does not include collection references

		return result;
	}	

	protected Object setExistingOrNewCopy(CallInfo ci) throws Exception {
		Object propertyTarget = null;

		// If this is a collection, then we need to skip creating copy if an existing persistent collection is found
		Object value = ci.getOutputFromParent();
		if(value != null) {
			// Existing embedded/collection objects need to be wrapped in a data object
			if(ci.getInputProperty() != null && 
					(ci.getInputProperty().isMany() || ((ExtendedType)ci.getInputProperty().getType()).isEmbedded()) && 
					!ExtendedDataObject.class.isAssignableFrom(value.getClass()))
				propertyTarget = createTarget(ci, value, null);
		}

		if(propertyTarget == null)
			propertyTarget =  createTarget(ci, null); // clear the existing id

		propertyTarget = setPropertyTarget(ci, propertyTarget);

		return propertyTarget;
	}	

	public Object getExistingTarget(CallInfo ci) {
		return ci.getOutputObjectCreator().getExistingDataObject(ci.getInput());
	}

	@Override
	public Object createTarget(CallInfo ci, Class<?> desiredClass) {
		return createTarget(ci, null, desiredClass);
	}

	@Override
	public Object createTarget(CallInfo ci, Object targetInstance, Class<?> desiredClass) {
		targetInstance = ClassUtil.getInstance(targetInstance);
		Node target = (Node) ci.getOutputObjectCreator().createTarget(ci, targetInstance, desiredClass);

		return target;
	}	

	protected Object setPropertyTarget(CallInfo ci, Object propertyTarget) throws PropertyMappingException {
		if(ci.getStage() != ProcessingStage.UPDATE) // Do not modify if not updating it
			return propertyTarget;

		// Set the target property value
		ci.linkOutputToParent(propertyTarget);			

		return propertyTarget;
	}		

	protected boolean shouldUnlink(CallInfo ci) {
		return true;	
	}	

	protected boolean isIdentifier(CallInfo ci) {
		if(ci.getInputProperty() == null)
			return false;

		ExtendedType sourceType = (ExtendedType) ((ExtendedDataObject) ci.getParent().getInput()).getType();
		if(sourceType.isEmbedded()) // embedded type
			return false;

		if(ci.getInputProperty().getName().equals( sourceType.getIdentifierProperty().getName()) )
			return true;

		return false;
	}

	protected void processNullValue(CallInfo ci) throws Exception {
		if(ci.getStage() != ProcessingStage.UPDATE) // Do not modify if not updating it
			return;

		// Set the null value - unlink the relationship from this side	
		ci.linkOutputToParent(null);				
	}	


	/**
	 * The Data objects for the source collection should have been created
	 * So use the source collection, retrieve the data objects and populate the
	 * target collection
	 * @param object
	 */
	public void cloneToMany(CallInfo callInfo) {
		if(callInfo.getStage() != ProcessingStage.UPDATE)
			return;

		Object object = ((ExtendedDataObject) callInfo.getInput()).getInstance();
		java.util.Collection sourceCollection = null;
		java.util.Collection targetCollection = null;
		Map sourceMap = null;
		Map targetMap = null;

		if(callInfo.getInputProperty().isSet() || callInfo.getInputProperty().isList()) {
			sourceCollection = (java.util.Collection) object;
			targetCollection = (java.util.Collection) ((ExtendedDataObject) callInfo.getOutput()).getInstance();

			for(Object source: sourceCollection) {
				ExtendedDataObject target = (ExtendedDataObject) callInfo.getOutputObjectCreator().getExistingDataObject(source);
				targetCollection.add(target.getInstance());
			}			
		} else if(callInfo.getInputProperty().isMap()) {
			sourceMap = (java.util.Map) object;
			targetMap = (java.util.Map) ((ExtendedDataObject) callInfo.getOutput()).getInstance();

			for(Object source: sourceMap.entrySet()) {
				Object targetKey = ((Map.Entry)source).getKey();
				ExtendedDataObject targetValue = (ExtendedDataObject) callInfo.getOutputObjectCreator().getExistingDataObject(((Map.Entry)source).getValue());
				targetMap.put(targetKey, targetValue);
			}				
		}
	}

	protected void processCollection(CallInfo callInfo) throws Exception {

		if(callInfo.isDataType())
			return;

		for (Object nextSource : ((ExtendedDataObject)callInfo.getInput()).getList(callInfo)) {
			CallInfo next = new CallInfo(nextSource, null, callInfo, null);
			if(next.isCascadable()) {
				next.setOutput(getExistingTarget(next));
				if( next.getOutput() == null )
					next.setOutput(createTarget(next, null));
				processValue(next);			
			} else
				next.setOutput(createTarget(next, ClassUtil.getInstance(nextSource), null));
		}
		cloneToMany(callInfo);
	}

	protected void customAddToMap(Object targetKey, CallInfo callInfo) {

		Object map                   = callInfo.getParent().getOutput();
		Object targetValue           = callInfo.getOutput();
		ExtendedProperty mapProperty = callInfo.getParent().getOutputProperty();		

		// check if there is a custom collection add
		Method customCollectionAdd = ClassUtil.getCustomAdd(mapProperty.getElementType().getInstanceClass(), mapProperty.getName());
		if(customCollectionAdd != null) {
			CollectionEvent event = new CollectionElement(callInfo, map, targetValue);
			mapProperty.setCustomAdd(map, event);
		} else
			((Map<Object, Object>)map).put(targetKey, targetValue);			
	}		
}
