/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model.operation;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import ws.ladder.model.CallInfo;
import ws.ladder.model.ExtendedDataObject;
import ws.ladder.model.Settings;
import ws.ladder.model.Type;
import ws.ladder.model.TypeMapper;
import ws.ladder.model.view.Query;
import ws.ladder.model.view.QueryBuilder;
import ws.ladder.model.view.ViewBranch;
import ws.ladder.service.DataAccessService;
import ws.ladder.util.ClassUtil;

public class QueryOperation extends AbstractOperation {

	private List<Object> result = new ArrayList<Object>();

	@Override
	protected void processPostLogic(CallInfo callInfo) {
		// Invariants do not need to be run since this operation does not modify data
	}
	
	private Type getNarrowedClass(DataAccessService das, Settings settings) {
		TypeMapper typeMapper = das.getTypeMapper();
		return das.getType(typeMapper.toReference(settings.getNarrowedClass()));
	}

	@Override
	public void execute(CallInfo callInfo) {
		ExtendedDataObject sourceEntity = (ExtendedDataObject) callInfo.getInput();
		DataAccessService das = sourceEntity.getObjectCreator().getDAS();
		QueryBuilder qb = das.getQueryBuilder();
		
		// Always use the REFERENCE type
		Type referenceType = (callInfo.getSettings().getNarrowedClass() == null) ? ((ExtendedDataObject) callInfo.getInput()).getReferenceType() : getNarrowedClass(das, callInfo.getSettings());
		ViewBranch aggregateView = callInfo.getSettings().getView().getEntityView( referenceType, callInfo.getSettings().doNarrow() );
		Map<ExtendedDataObject, Object> uniqueList = new LinkedHashMap<ExtendedDataObject, Object>();	
		
		// Put in loop if there are sub-branches
		if(aggregateView.getSubBranches().size() > 1) {
			for(ViewBranch branch: aggregateView.getSubBranches())
				executeBranch(branch, uniqueList, qb, callInfo);
		} else
			executeBranch(aggregateView, uniqueList, qb, callInfo);
		
		// Do a second pass
		for(ExtendedDataObject root: uniqueList.keySet())
			if(root.getContainer() == null) // Only add root objects
				result.add(root);
	}
	
	private Query createQuery(CallInfo callInfo, QueryBuilder qb) {
		Map<String, Object> mutableFilters = new HashMap<String, Object>(callInfo.getSettings().getFilters());
		String queryString = qb.constructQuery(callInfo, mutableFilters);
		
		Query query = callInfo.getInputObjectCreator().getPersistenceOrchestrator().getQuery(queryString, qb.isSQL());

		qb.postProcess(callInfo, query, mutableFilters);
		
		return query;				
	}
	
	private void executeBranch(ViewBranch branch, Map<ExtendedDataObject, Object> uniqueList, QueryBuilder qb, CallInfo callInfo) {
		qb.init((ExtendedDataObject) callInfo.getInput(), branch, callInfo.getSettings().getAdditionalFilters());	
		Query query = createQuery(callInfo, qb);
		
		try {
			// club all the results relevant to the same entity
			// add an id attribute for the mail entity. Add an owner attribute for each collection property referenced.
			// adjust the properties and for every new attribute added (id or owner) create a filler/dummy property column in the view
			if(branch.isCrossAggregate() && !callInfo.getSettings().permitCrossAggregate())
				throw new RuntimeException("The view crosses aggregate boundary, this could be a security risk. If this is intentional then permit this by modifying the settings");
			
			for(Object obj: query.getResultList(branch)) {
				ExtendedDataObject newRootObject = branch.getQueryRoot(obj, (ExtendedDataObject) callInfo.getOutput());

				if(ClassUtil.getDimensionCount(obj) == 1) {
					branch.normalize(newRootObject, (Object[])obj);
					if(newRootObject.getContainer() == null && !uniqueList.containsKey(newRootObject)) // Only add root objects
						uniqueList.put(newRootObject, null);
				}
			}
		} catch (Exception e) {
			throw ClassUtil.wrapRun(e);
		}		
	}

	@Override
	public Object getResult() {
		return result;
	}		
}
