/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model.view;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import ws.ladder.model.CallInfo;
import ws.ladder.model.ExtendedDataObject;
import ws.ladder.model.ExtendedType;
import ws.ladder.service.QueryCapability;

/**
 * QueryBuilder needs to handle the following expressions

First Phase -- Restrict to what is needed by CSA 

CSA has the following filters
-----------------------------
ilike(param, %1)
in(param, %1)
idEq(%1)
ge(%1)
asc(%1)

e.g.

ilike(name, "%xyz%")
in("id", String[])
idEq(String)
ge("createdOn", long)
ge("updatedOn", long)
asc("name")

Modify view to support filters
------------------------------

  FILTERS
  ilike(name, ":name")            // lower(name) LIKE :name
  in(state, ":state"              // name IN(:state)
  equal(ownedBy.name, ":owner")   // ownedBy.name = :owner 
  ge(createdOn, ":createdSince")  // createdOn > :createdSince
  ge(updatedOn, ":updatedSince"); // updaetdOn > :updatedSince
  asc(name)                       // ORDER BY name  also honor OrderBy annotation in model if no ORDER BY clause is provided
  between(updatedOn, ":updatedFrom", ":updatedTo")

 *
 */
public class QueryBuilder {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());

	private static final String SELECT_CLAUSE = "SELECT ";
	private static final String COMMA_DELIMITER = ", ";
	public  static final String ID_PARAMETER_NAME = "id_";

	private ExtendedDataObject entity;
	private ViewBranch         view;
	private List<Filter>       additionalFilters;
	private boolean            SQL;
	private boolean            addIdentifier;

	public ExtendedDataObject getEntity() {
		return entity;
	}	

	public boolean doAddIdentifier() {
		return addIdentifier;
	}

	public void setAddIdentifier(boolean addIdentifier) {
		this.addIdentifier = addIdentifier;
	}

	public boolean isSQL() {
		return SQL;
	}

	public void init(ExtendedDataObject entity, ViewBranch view, List<Filter> additionalFilters) {
		this.entity = entity;
		this.view = view;

		this.additionalFilters = new ArrayList<Filter>();
		for(Filter filter: additionalFilters) {
			Filter narrowedFilter = filter.narrow();
			this.additionalFilters.add(narrowedFilter);	
		}	
		this.additionalFilters.addAll(view.getFilters());		
		view.normalizeFilters(this.additionalFilters, getQueryCapability());

		if(entity.getIdentifierValue() != null)
			addIdentifier = true;
	}

	private QueryCapability getQueryCapability()  {
		return entity.getObjectCreator().getPersistenceOrchestrator().getQueryCapability();
	}

	public String constructQuery(CallInfo callInfo, Map<String, Object> filters) {
		//		First check for a SQL query 
		//		When retrieving the QueryBuilder instance, registered SQL queries are given preference over HQL/JPQL queries.
		//
		//		It is the responsibility of the user to ensure that the registered SQL is ANSI compliant 
		//		and can work with all the databases that the user uses.

		if(view.getContentView() != null && view.getContentView().getNativeQuery() != null) {
			SQL = true;
			String queryString = view.getContentView().getNativeQuery().getQueryString();
			if(doAddIdentifier()) {
				String idClause = view.getContentView().getNativeQuery().getIdentifierClause();
				if((idClause == null || idClause.trim().equals("")))
					throw new RuntimeException("Identifier clause expected in native query");

				queryString = queryString + " " + idClause + " = :" + ID_PARAMETER_NAME;
			}

			return queryString.replaceAll("[\n\r]", "");
		}

		StringBuffer HQL = new StringBuffer(constructHQL(callInfo));
		HQL.append(buildWhereClause(filters));
		HQL.append(buildOrderClause());

		return HQL.toString();		
	}

	protected String constructHQL(CallInfo callInfo) {
		Map<String, ColumnMeta> meta = view.getAugmentedAttributes();

		StringBuffer HQL = new StringBuffer(SELECT_CLAUSE);

		int position = 0;
		for(ColumnMeta columnMeta: meta.values()) {
			if(!columnMeta.getViewProperty().doFetch()) // Should the property not be fetched
				continue;

			columnMeta.setPosition(position++);			
			if(!SELECT_CLAUSE.equals(HQL.toString()))
				HQL.append(COMMA_DELIMITER);
			String columnString = columnMeta.getQueryString(getQueryCapability());
			HQL.append(columnString);
		}

		Class<?> entityClass = (callInfo.getSettings().getNarrowedClass() == null) ? entity.getReferenceType().getInstanceClass() : callInfo.getSettings().getNarrowedClass();
		HQL.append(" FROM " + entityClass.getName() + " AS " + view.getViewProperty(ViewProperty.ROOT_PROPERTY_NAME).getAlias());

		// if more than one collection property is specified then we would get a severe performance hit due to cartesian product
		// This is solved by breaking the query intelligently. 
		for(ViewProperty viewProperty: view.getAliasedItems()) {
			ViewProperty parentView = viewProperty.getParent();
			if(parentView == null)
				continue; // the root needs to be skipped
			HQL.append(" LEFT OUTER JOIN " + viewProperty.getNormalizedName() + " AS " + viewProperty.getAlias());		
		}

		if(view.getContentView() != null && view.getContentView().getJoin() != null && view.getContentView().getJoin().getEntity() != null)
			HQL.append(", " + view.getContentView().getJoin().getEntity());		

		return HQL.toString();
	}	

	protected String buildOrderClause() {
		StringBuffer result = new StringBuffer();

		Map<Integer, String> orderClauses = new TreeMap<Integer, String>();
		for(ColumnMeta columnMeta: view.getAugmentedAttributes().values()) {
			if(columnMeta.getAttributePath().endsWith(ViewProperty.LIST_INDEX_ATTRIBUTE)) {
				String columnString = columnMeta.getQueryString(getQueryCapability());
				orderClauses.put(columnMeta.getAttributeLevel(), columnString);
			}
		}

		// filter order clauses
		StringBuffer filterOrderBy = new StringBuffer();
		for(Filter filter: this.additionalFilters) {
			if(filter.isOrderBy()) {
				if(filterOrderBy.length() > 0)
					filterOrderBy.append(COMMA_DELIMITER);
				filterOrderBy.append(filter.getQueryString());
			}
		}

		// Append any ORDER BY clauses
		if(orderClauses.size() > 0 || filterOrderBy.length() > 0) {
			for(String indexColumn: orderClauses.values())
				result.append( (result.length() > 0) ? COMMA_DELIMITER + indexColumn : indexColumn );

			if(filterOrderBy.length() > 0) 
				result.append((result.length() > 0) ? COMMA_DELIMITER + filterOrderBy : filterOrderBy);

			result.insert(0, " ORDER BY ");			
		}		

		return result.toString();
	}

	protected String getNormalizedName(String propertyPath) {
		ViewProperty viewProperty = view.getViewProperty(ViewProperty.qualifyProperty(propertyPath));
		return viewProperty.getNormalizedName();
	}	

	protected String buildWhereClause(Map<String, Object> filters) {
		StringBuffer result = new StringBuffer();
		Map<String, Object> originalFilters = new HashMap<String, Object>(filters);
		filters.clear();

		Map<String, Parameter> parameterMap = new HashMap<String, Parameter>();
		for(Parameter parameter: view.getParameter())
			parameterMap.put(parameter.name, parameter);

		// Initialize the filterbypath map
		for(Filter filter: this.additionalFilters) {
			if(filter.isOrderBy()) // Will be handled in the ORDER BY clause
				continue;

			if(!filter.updateFilters(originalFilters, filters, parameterMap))
				continue;
			addWhereStep(result);

			result.append(filter.getQueryString());
		}

		if(addIdentifier) {
			addWhereStep(result);

			result.append( getNormalizedName(((ExtendedType)entity.getReferenceType()).getIdentifierProperty().getName()) );
			result.append( " = :" + ID_PARAMETER_NAME);			
		}

		// Set parameters referred from native query or join clause
		if(view.getContentView() != null && view.getContentView().getJoin() != null) {
			// Expression
			addWhereStep(result);

			for(Parameter parameter: view.getContentView().getJoin().getParameter()) {		
				if(originalFilters.containsKey(parameter.name)) {
					filters.put(parameter.name, originalFilters.get(parameter.name));
					result.append(parameter.expression);
				}
			}
		}

		return result.toString();
	}	

	private void addWhereStep(StringBuffer result) {
		if(result.length() > 0)
			result.append(" AND ");		
		else
			result.append(" WHERE ");		
	}

	public void postProcess(CallInfo callInfo, Query query, Map<String, Object> filters) {
		for(String parameterName: filters.keySet())
			query.setParameter(parameterName, filters.get(parameterName));

		if(doAddIdentifier() && query.hasParameter(ID_PARAMETER_NAME))
			query.setParameter(ID_PARAMETER_NAME, getEntity().getIdentifierValue());			

		// pagination
		if(callInfo.getSettings().getOffset() != null)
			query.setFirstResult(callInfo.getSettings().getOffset());
		if(callInfo.getSettings().getLimit() != null)
			query.setMaxResults(callInfo.getSettings().getLimit());		
	}
}
