/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.service;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import ws.ladder.model.AbstractType;
import ws.ladder.model.DerivedType;
import ws.ladder.model.ExtendedType;
import ws.ladder.model.SimpleType;
import ws.ladder.model.Type;
import ws.ladder.model.TypeMapper;
import ws.ladder.model.view.QueryBuilder;

public abstract class AbstractDataAccessService implements DataAccessService {

	protected TypeMapper        typeMapper;	
	protected Map<String, Type> types = new ConcurrentHashMap<String, Type>();
	protected Map<String, Type> derivedTypes = new ConcurrentHashMap<String, Type>();	
	protected DASFactory        dasFactory;
	
	public AbstractDataAccessService(DASFactory factory) {
		this.dasFactory = factory;
	}

	protected void addDerivedType(String className, Type type) {
		addType(className, type, derivedTypes);
	}	

	protected void addType(String className, Type type) {
		addType(className, type, types);
	}

	protected void addType(String className, Type type, Map<String, Type> typeMap) {
		typeMap.put(className, type);

		if(ExtendedType.class.isAssignableFrom(type.getClass())) {
			String entityName = ((ExtendedType)type).getEntityName();
			if(entityName != null && !className.equals(entityName)) {
				if(typeMap.containsKey(entityName)) {
					if(!className.equals(typeMap.get(entityName).getInstanceClass().getName()))
						throw new RuntimeException("Type " + typeMap.get(entityName).getName() + " already exists for entityName: " + entityName);
				} else
					typeMap.put(entityName, type);
			}
		}
	}
	
	protected void defineSubtypes(){
		List<ExtendedType> entityTypes = new ArrayList<ExtendedType>();
		for(Type type: getUniqueTypes()) {
			if(!ExtendedType.class.isAssignableFrom(type.getClass()))
				continue;
			
			ExtendedType extendedType = (ExtendedType) type;
			entityTypes.add(extendedType);
		}
		
		for(ExtendedType type: entityTypes)
			type.defineSubtypes(entityTypes);
	}

	@Override
	public Type getDerivedType(Class<?> clazz) {
		Type result = derivedTypes.get(clazz.getName());

		// create a Type object for this class
		if (result == null) {
			result = new SimpleType(clazz);
			addDerivedType(clazz.getName(), result);
		}

		return result;
	}	

	@Override
	public Type getType(String name) {
		return types.get(name);
	}

	@Override
	public Type getDerivedType(String name) {
		return derivedTypes.get(name);
	}	

	@Override
	public TypeMapper getTypeMapper() {
		return typeMapper;
	}	

	@Override
	public List<Type> getTypes() {
		return new ArrayList<Type>(getUniqueTypes());
	}	

	/**
	 * This method uses the persistence mechanism to do any post-processing activity
	 * @param newInstance
	 * @return
	 */

	@Override
	public void postProcess(Object newInstance) {
		if(newInstance == null)
			return;

		// Since we are manually creating the instance we need to autowire any dependencies
		dasFactory.injectDependencies(newInstance, null);
	}

	protected void postProcess() {
		initPositionProperty();
		
		initDerived();

		initRootType();
	}
	
	private Set<Type> getUniqueTypes() {
		return new HashSet<Type>(types.values());
	}
	
	private Set<Type> getUniqueDerivedTypes() {
		return new HashSet<Type>(derivedTypes.values());
	}

	private void initRootType() {
		for (Type type : getUniqueTypes()) {
			if (AbstractType.class.isAssignableFrom(type.getClass())) {
				((AbstractType)type).initRootEntityType(this);
			}
		}
		for (Type type : getUniqueDerivedTypes()) {
			if (AbstractType.class.isAssignableFrom(type.getClass())) {
				((AbstractType)type).initRootEntityType(this);
			}
		}			
	}
	
	private void initPositionProperty() {
		for (Type type : getUniqueTypes()) {
			if (AbstractType.class.isAssignableFrom(type.getClass())) {
				((AbstractType)type).initPositionProperty();
			}
		}
	}

	protected void initDerived() {
		for(Type type: getUniqueTypes()) {
			if(SimpleType.class.isAssignableFrom(type.getClass()))
				continue;
			Class<?> derivedClass = typeMapper.toDerived(type.getInstanceClass());
			if(derivedClass != null) {
				Type derived = new DerivedType((ExtendedType) type, derivedClass);
				derivedTypes.put(derived.getName(), derived);
			}
		}

		// init the derived properties
		for (Type type : getUniqueDerivedTypes()) {
			if (DerivedType.class.isAssignableFrom(type.getClass())) {
				DerivedType derivedType = (DerivedType) type;
				derivedType.setProperty(this);
			}
		}		

		for (Type type : getUniqueDerivedTypes()) {
			if (DerivedType.class.isAssignableFrom(type.getClass())) {
				DerivedType derivedType = (DerivedType) type;
				derivedType.setOpposite(this);
			}
		}			
	}

	public QueryBuilder getQueryBuilder() {
		return new QueryBuilder();
	}
}