/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.service;

import java.util.ArrayList;
import java.util.List;

import javax.annotation.PostConstruct;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.stereotype.Component;

import ws.ladder.core.Interceptor;
import ws.ladder.model.AggregateAction;
import ws.ladder.model.DefaultTypeMapper;
import ws.ladder.model.DefaultTypeNarrower;
import ws.ladder.model.ExtendedDataObject;
import ws.ladder.model.ExtendedType;
import ws.ladder.model.MapperDirection;
import ws.ladder.model.Settings;
import ws.ladder.model.TypeMapper;
import ws.ladder.model.TypeNarrower;
import ws.ladder.model.custom.AssociationStrategy;
import ws.ladder.model.custom.DefaultAssociationStrategy;
import ws.ladder.model.view.ContentViewFactory;
import ws.ladder.util.AccessType;
import ws.ladder.util.ObjectCreator;

@Component
public class AggregateManager {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());

	protected DASFactory              dasFactory;	
	protected List<String>            viewFiles;           // A list of files to read the view information from
	protected MetaModel               metaModel;           // Meta model exposed to the user
	protected boolean                 crossAggregate;      // A system wide setting to allow cross aggregate access
	protected boolean                 autoFlushNative;     // Flag to indicate if a flush should be issued before a native query is executed
	protected Interceptor             interceptor;         // A user provided interceptor to be notified of framework events
	protected AssociationStrategy     associationStrategy;
	protected boolean                 generateMetaModel;   // Generated metamodel on startup. This allows content integrators to find out the aggregate properties to be used in the views.
	protected TypeMapper              typeMapper;          // Custom type mapper to map between differnt types. NOTE: The user is restricted from changing the path specified in the MetaModel.
	// A truly custom solution is to provide a map between the meta model path and the custom path.
	// Configured with the DefaultTypeMapper
	protected AccessType              accessType;
	protected TypeNarrower            typeNarrower;
	protected PersistenceOrchestrator persistenceOrchestrator;

	@PostConstruct
	protected void init() {
		// Load the default view file
		(new ContentViewFactory()).load();

		// Load the configured view files
		syncViews();

		if(associationStrategy == null)
			associationStrategy = new DefaultAssociationStrategy();

		if(typeMapper == null)
			typeMapper = new DefaultTypeMapper();

		if(typeNarrower == null) {
			typeNarrower = new DefaultTypeNarrower();
			typeNarrower.setAggregateManager(this);
		}

		if(dasFactory != null) {
			metaModel = new MetaModel(dasFactory);
			dasFactory.setAggregateManager(this);
		} else
			logger.error("DASFactory instance is not set on the AggregateManager instance with the default name.");
	}

	public DASFactory getDasFactory() {
		return dasFactory;
	}

	public void setDasFactory(DASFactory dasFactory) {
		this.dasFactory = dasFactory;
	}

	public MetaModel getMetaModel() {
		return metaModel;
	}

	public void setMetaModel(MetaModel metaModel) {
		this.metaModel = metaModel;
	}

	public void syncViews() {
		if(viewFiles == null)
			return;

		// Load the views
		ContentViewFactory viewFactory = new ContentViewFactory();

		for(String viewFile: viewFiles)
			viewFactory.load(viewFile);		
	}

	public TypeNarrower getTypeNarrower() {
		return typeNarrower;
	}

	public void setTypeNarrower(TypeNarrower typeNarrower) {
		this.typeNarrower = typeNarrower;
		typeNarrower.setAggregateManager(this);
	}	

	public AssociationStrategy getAssociationStrategy() {
		return associationStrategy;
	}

	public void setAssociationStrategy(AssociationStrategy associationStrategy) {
		this.associationStrategy = associationStrategy;
	}	

	public AccessType getAccessType() {
		return accessType;
	}

	public void setAccessType(AccessType accessType) {
		this.accessType = accessType;
	}	

	public List<String> getViewFiles() {
		return viewFiles;
	}

	public void setViewFiles(List<String> viewFiles) {
		this.viewFiles = viewFiles;
	}

	public boolean isCrossAggregate() {
		return crossAggregate;
	}

	public void setCrossAggregate(boolean crossAggregate) {
		this.crossAggregate = crossAggregate;
	}

	public boolean isAutoFlushNative() {
		return autoFlushNative;
	}

	public void setAutoFlushNative(boolean autoFlushNative) {
		this.autoFlushNative = autoFlushNative;
	}

	public Interceptor getInterceptor() {
		return interceptor;
	}

	public void setInterceptor(Interceptor interceptor) {
		this.interceptor = interceptor;
	}

	public boolean isGenerateMetaModel() {
		return generateMetaModel;
	}

	public void setGenerateMetaModel(boolean generateMetaModel) {
		this.generateMetaModel = generateMetaModel;
	}

	public TypeMapper getTypeMapper() {
		return typeMapper;
	}

	public void setTypeMapper(TypeMapper customTypeMapper) {
		this.typeMapper = customTypeMapper;
	}	

	public PersistenceOrchestrator getPersistenceOrchestrator() {
		return persistenceOrchestrator;
	}

	public void setPersistenceOrchestrator(
			PersistenceOrchestrator persistenceOrchestrator) {
		this.persistenceOrchestrator = persistenceOrchestrator;
	}	

	public DataAccessService getDAS() {
		return dasFactory.create();
	}

	public Settings getSettings() {
		Settings result = new Settings();
		result.setAssociationStrategy(associationStrategy);

		return result;
	}

	private void checkAndSet(Settings settings) {
		if(persistenceOrchestrator == null)
			persistenceOrchestrator = dasFactory.getPersistenceOrchestrator();		
		if(settings.getAssociationStrategy() == null)
			settings.setAssociationStrategy(associationStrategy);
	}

	private void reset() {
		persistenceOrchestrator = null;
	}

	public Object clone(Object entity, Class<?> desiredSubClass, Settings settings) {
		checkAndSet(settings);
		DataAccessService das = getDAS();

		ObjectCreator oc = new ObjectCreator(das, getPersistenceOrchestrator(), MapperDirection.REFERENCETOREFERENCE);
		ExtendedDataObject from = oc.createDataObject(entity, (ExtendedType) das.getType(entity.getClass()), null, null);
		ExtendedDataObject to = (ExtendedDataObject) from.clone(settings);		

		reset();
		return to.getInstance();
	}

	public void delete(Object entity) {
		// TODO: 
	}	

	public List<?> query(Object entity, Settings settings) {
		checkAndSet(settings);

		DataAccessService das = getDAS();
		if(settings.doPreFlush())
			getPersistenceOrchestrator().flush();

		MapperDirection direction = MapperDirection.DERIVEDTODERIVED;
		if(das.getTypeMapper().isReference(entity.getClass()))
			direction = MapperDirection.REFERENCETODERIVED;

		ObjectCreator oc = new ObjectCreator(das, getPersistenceOrchestrator(), direction);
		ExtendedDataObject from = oc.createDataObject(entity, (ExtendedType) oc.getSourceType(entity.getClass()), null, null);

		// Get the narrowed class
		settings.initNarrowClass(getTypeNarrower(), entity, typeMapper);

		List<?> dataObjects = from.query(settings);
		List<Object> result = new ArrayList<Object>();
		for(Object obj: dataObjects) 
			result.add( ((ExtendedDataObject)obj).getInstance() );

		reset();
		return result;
	}

	public Object read(Object entity, Settings settings) {
		checkAndSet(settings);		

		ObjectCreator oc = new ObjectCreator( getDAS(), getPersistenceOrchestrator(), MapperDirection.REFERENCETODERIVED);
		ExtendedDataObject from = oc.createDataObject(entity, oc.getSourceType(entity.getClass()), null, null);
		if(!from.isReferenceType())
			from = from.load(getSettings());  // Get the persistent object

		//  perform read on it
		ExtendedDataObject to = (ExtendedDataObject) from.read(settings);		

		reset();
		return to.getInstance();
	} 	

	public Object create(Object entity, Settings settings) {
		checkAndSet(settings);		

		ObjectCreator oc = new ObjectCreator(getDAS(), getPersistenceOrchestrator(), MapperDirection.DERIVEDTOREFERENCE);
		ExtendedDataObject from = oc.createDataObject(entity, (ExtendedType) oc.getSourceType(entity.getClass()), null, null);
		ExtendedDataObject to = (ExtendedDataObject) from.create(settings);					

		reset();
		return to.getInstance();		
	}

	public Object update(Object entity, Settings settings) {
		checkAndSet(settings);		

		if(settings.getAction() != AggregateAction.MERGE && settings.getAction() != AggregateAction.UPDATE)
			throw new IllegalStateException("The default action should either by UPDATE or MERGE");

		ObjectCreator oc = new ObjectCreator(getDAS(), getPersistenceOrchestrator(), MapperDirection.DERIVEDTOREFERENCE);
		ExtendedDataObject from = oc.createDataObject(entity, (ExtendedType) oc.getSourceType(entity.getClass()), null, null);
		ExtendedDataObject to = (ExtendedDataObject) from.update(settings);	

		if(settings.doPostFlush())
			getPersistenceOrchestrator().flush();

		reset();
		return to.getInstance();
	}	

	public void linkBackPointer(Object entity) {
		ObjectCreator oc = new ObjectCreator(getDAS(), getPersistenceOrchestrator(), MapperDirection.DERIVEDTODERIVED);
		ExtendedDataObject dataObject = oc.createDataObject(entity, (ExtendedType) oc.getSourceType(entity.getClass()), null, null);
		oc.setShare(true);		
		dataObject.createAggregate();
		dataObject.linkBackPointer();
	}
}
