/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.service;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.hibernate.SessionFactory;
import org.hibernate.cfg.Configuration;
import org.hibernate.mapping.PersistentClass;
import org.springframework.beans.factory.annotation.Autowired;

import ws.ladder.model.HibernateType;
import ws.ladder.model.SimpleType;
import ws.ladder.model.Type;
import ws.ladder.model.TypeMapper;
import ws.ladder.util.AccessType;
import ws.ladder.util.HibernateUtil;

/**
 * This class is part of the Data Access Service framework
 * 
 * @author Dilip Dalton
 * 
 */
public class HibernateDAS extends AbstractDataAccessService {

	private static final Logger logger = LogManager.getLogger(new Exception()
	.getStackTrace()[0].getClassName());
	
	@Autowired(required = false)
	protected Configuration configuration;	

	@Autowired
	protected SessionFactory sessionFactory;	
	
	public HibernateDAS(TypeMapper typeMapper, DASFactory dasFactory) {
		super(dasFactory);
		this.typeMapper = typeMapper;
	}
	
	public SessionFactory getSessionFactory() {
		return sessionFactory;
	}
	
	public Configuration getConfiguration() {
		return configuration;
	}

	public void setConfiguration(Configuration configuration) {
		this.configuration = configuration;
	}	

	@Override
	public Type getType(Class<?> clazz) {
		Type result = super.getType(clazz.getName());

		// create a Type object for this class
		if (result == null) {
			result = new SimpleType(clazz);
			addType(clazz.getName(), result);
		}

		return result;
	}

	public void define() {
		Configuration conf = getConfiguration();
		Iterator<PersistentClass> classMappings = conf.getClassMappings();

		logger.info("Getting the list of hibernate mapped classes");
		while (classMappings.hasNext()) {
			PersistentClass classMapping = (PersistentClass) classMappings
					.next();
			logger.debug("     Adding hibernate persisted class: "
					+ classMapping.getClassName());
			defineTypes(classMapping);
		}

		// Set the base types
		setBaseTypes();

		// Define the properties for the Types
		// This will end up defining the simple types
		defineProperties();
		
		postProcess();
	}

	protected void defineTypes(PersistentClass classMapping) {
		HibernateType dataType = new HibernateType( HibernateUtil.getEntityType(sessionFactory, classMapping.getEntityName()),
				classMapping);
		
		logger.debug("Defined data type: " + dataType.getName());
		addType(dataType.getName(), dataType);
		
		for(Type type: dataType.getEmbeddableTypes()) {
			addType(type.getName(), type);
		}
		
		defineSubtypes();
	}

	protected void defineProperties() {
		for (Type type : types.values()) {
			if (HibernateType.class.isAssignableFrom(type.getClass())) {
				HibernateType hibernateType = (HibernateType) type;
				hibernateType.setProperty(this);
			}
		}

		// Link the bi-directional relationship between the properties
		for (Type type : types.values()) {
			if (HibernateType.class.isAssignableFrom(type.getClass())) {
				HibernateType hibernateType = (HibernateType) type;
				hibernateType.setOpposite(this);
			}
		}
	}

	protected void setBaseTypes() {
		for (Type type : types.values()) {
			if (HibernateType.class.isAssignableFrom(type.getClass())) {
				HibernateType hibernateType = (HibernateType) type;
				if(hibernateType.getHibernateType().isComponentType())
					continue;

				List<Type> baseTypes = new ArrayList<Type>();
				PersistentClass base = ((PersistentClass)hibernateType.getHibernateClass()).getSuperclass();

				if (base != null) {
					Type baseType = types.get(base.getEntityName());
					if (baseType != null)
						baseTypes.add(baseType);	
				}
				hibernateType.setBaseType(baseTypes);
			}
		}
	}

	@Override
	public List<String> getAggregateList() {
		List<String> result = new ArrayList<String>();
		
		Configuration conf = getConfiguration();
		Iterator<PersistentClass> classMappings = conf.getClassMappings();

		logger.info("Getting the list of hibernate mapped classes");
		while (classMappings.hasNext()) {
			PersistentClass classMapping = (PersistentClass) classMappings
					.next();
			result.add(classMapping.getEntityName());
		}
		
		return result;
	}

	@Override
	public AccessType getAccessType() {
		return AccessType.HIBERNATE;
	}
}