/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.service;

import java.io.Serializable;
import java.util.List;
import java.util.Map;

import org.hibernate.Criteria;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.criterion.Restrictions;
import org.springframework.beans.factory.annotation.Autowired;

import ws.ladder.model.Type;
import ws.ladder.model.view.HibernateQuery;
import ws.ladder.model.view.Query;
import ws.ladder.util.HibernateUtil;

public class HibernatePersistenceOrchestrator extends AbstractPersistenceOrchestrator {

	@Autowired
	private SessionFactory sessionFactory;    

	protected Session getSession() {
		if (this.sessionFactory.getCurrentSession() == null)
			throw new IllegalStateException("Session has not been set on DAO before usage");
		return this.sessionFactory.getCurrentSession();
	}    	

	/**
	 * Save the entity in the persistence store
	 * 
	 * @param entity
	 * @return
	 */
	@Override
	public void saveOrUpdate(Object entity) {
		getSession().saveOrUpdate(entity);		
	}

	/**
	 * Delete the entity from the persistence store
	 * 
	 * @param entity
	 */
	@Override
	public void delete(Object entity) {
		getSession().delete(entity);		
	}

	@Override
	public void flush() {
		getSession().flush();
	}

	@Override
	public Object findById(Class<?> persistentClass, Object id) {
		return getSession().get(persistentClass, (Serializable) id);
	}

	@Override
	public Object findByProperty(Type type, Map<String, String> propertyValues) {
		Class<?> persistentClass = type.getInstanceClass();
		Criteria crit = getSession().createCriteria(persistentClass);

		for(String property: propertyValues.keySet())
			crit.add(Restrictions.eq(property, propertyValues.get(property)));

		List<Object> resultList = crit.list();
		if (resultList != null && !resultList.isEmpty()) {
			return resultList.get(0);
		}
		return null;		
	}

	@Override
	public QueryCapability getQueryCapability() {
		return new JPAQueryCapability();
	}

	@Override
	public Query getQuery(String queryString, boolean isSQL) {
		
		Query result = null;
		if(isSQL) {
			result = new HibernateQuery(HibernateUtil.getCurrentSession(sessionFactory).createSQLQuery(queryString));
		} else
			result = new HibernateQuery(HibernateUtil.getCurrentSession(sessionFactory).createQuery(queryString));

		return result;
	}	
}