/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.service;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.persistence.metamodel.EntityType;
import javax.persistence.metamodel.Metamodel;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import ws.ladder.model.JPAType;
import ws.ladder.model.SimpleType;
import ws.ladder.model.Type;
import ws.ladder.model.TypeMapper;
import ws.ladder.util.AccessType;

/**
 * This class is part of the Data Access Service framework 
 * @author Dilip Dalton
 *
 */
public class JPADataAccessService extends AbstractDataAccessService {

	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());	

	@PersistenceContext
	private EntityManager entityManager;	
	
	public JPADataAccessService(TypeMapper typeMapper, DASFactory dasFactory) {
		super(dasFactory);
		this.typeMapper = typeMapper;
	}
	
	@Override
	public Type getType(Class<?> clazz) {
		Type result = super.getType(clazz.getName());		

		// create a Type object for this class
		if(result == null) {
			result = new SimpleType(clazz);
			addType(clazz.getName(), result);
		}

		return result;
	}	

	public void define() {

		Metamodel metaModel = entityManager.getMetamodel();
		Set<EntityType<?>> classMappings = metaModel.getEntities();

		logger.info("Getting the list of hibernate mapped classes");  		
		for(EntityType<?> classMapping: classMappings){ 
			logger.debug("     Adding hibernate persisted class: " + classMapping.getName());
			defineTypes(classMapping);
		}		

		// Set the base types
		setBaseTypes();		

		// Define the properties for the Types 
		// This will end up defining the simple types
		defineProperties();	
		
		postProcess();			
	}

	protected void defineTypes(EntityType<?> classMapping) {
		JPAType dataType = new JPAType(classMapping);
		logger.debug("Defined data type: " + dataType.getName());
		addType(classMapping.getJavaType().getName(), dataType);
		
		for(Type type: dataType.getEmbeddableTypes()) {
			addType(type.getName(), type);
		}		
	}

	protected void defineProperties() {
		for(Type type: types.values()) {
			if(JPAType.class.isAssignableFrom(type.getClass())) {
				JPAType jPAType = (JPAType) type;
				jPAType.setProperty(this);
			}
		}

		// Link the bi-directional relationship between the properties
		for(Type type: types.values()) {
			if(JPAType.class.isAssignableFrom(type.getClass())) {
				JPAType jPAType = (JPAType) type;
				jPAType.setOpposite(this);
			}			
		}		
	}	

	protected void setBaseTypes() {
		for(Type type: types.values()) {
			if(JPAType.class.isAssignableFrom(type.getClass())) {
				
				JPAType jPAType = (JPAType) type;
				
				if(jPAType.getEntityType().getPersistenceType() == javax.persistence.metamodel.Type.PersistenceType.EMBEDDABLE)
					continue;				

				List<Type> baseTypes = new ArrayList<Type>();
				Class<?> base = jPAType.getEntityType().getJavaType().getSuperclass();

				if(base != null) {
					Type baseType = types.get(base.getName());
					if(baseType != null) 
						baseTypes.add(baseType);
				}
				jPAType.setBaseType(baseTypes);
			}
		}
	}

	@Override
	public List<String> getAggregateList() {
		List<String> result = new ArrayList<String>();

		Metamodel metaModel = entityManager.getMetamodel();
		Set<EntityType<?>> classMappings = metaModel.getEntities();

		for(EntityType<?> classMapping: classMappings){ 
			defineTypes(classMapping);
			result.add(classMapping.getJavaType().getName());
		}		
		
		return result;
	}

	@Override
	public AccessType getAccessType() {
		return AccessType.JPA;
	}
}