/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.util.btree;

import java.util.HashMap;
import java.util.Map;

public abstract class AbstractStore implements Store {
	public static final int MIN_BLOCK_SIZE = 512;	

	protected int             blockSize;
	protected int             entrySize;
	protected IndexTypeEnum   type;
	protected Integer         minimumDegree;
	protected long            rootAddress;	
	protected Map<Long, Node> cachedNodes = new HashMap<Long, Node>();	

	public AbstractStore(int blockSize, int entrySize, IndexTypeEnum type) {
		this.blockSize = blockSize;
		this.entrySize = entrySize;
		this.type = type;
	}

	@Override
	public int getBlockSize() {
		return this.blockSize;
	}

	@Override
	public int getMinimumDegree() {
		/* 
		 * t is the minimum degree
		 * 
		 * - t >= 2
		 * - Every node other than the root must have at least t-1 keys. Every
		 *   internal node other than the root thus has at least t children.
		 *   If the tree is nonempty, the root must have at least one key.
		 * - Every node may contain at most 2t-1 keys. Therefore, an internal node
		 *   may have at most 2t children. We say that a node is full if it 
		 *   contains exactly 2t-1 keys.
		 */

		if(minimumDegree == null) {
			int dataLen = this.blockSize - Node.NODE_HEADER_LEN;
			int t = ((dataLen-Node.CHILD_PTR_LEN)/(Node.CHILD_PTR_LEN+entrySize) + 1)/2;
			
			if(t < 2)
				throw new RuntimeException("The minimum degree has to be >= 2");
			
			minimumDegree = t;
		}

		return minimumDegree;
	}	
	
	@Override
	public Node get(long address) {
		return cachedNodes.get(address);
	}	
	
	@Override
	public long getRootAddress() {
		return rootAddress;
	}	
}