/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.util.btree;

import java.io.IOException;
import java.util.List;

/**
 * Implementation of BTree using RandomAccessFile implementation. The algorithm
 * is based off Introduction to Algorithms by CLRS.
 **/
public class BTree {

	public static final int MIN_NUM_KEYS = 4;

	private volatile Node            root;      // the root of the B-tree is always in main memory
	private          Store           store;     // used for persistence of the index
	private          IndexType       indexType; // Object used to work with the index entries

	public BTree(Store store) {
		this.store = store;
		this.indexType = store.getIndexType();
		this.root = getRoot();
	}

	private int getMinimumDegree() {
		return store.getMinimumDegree();
	}

	private void setRoot(Node r) {
		root = r;
		store.setRoot(root);
	}

	public IndexType getIndexType() {
		return indexType;
	}

	public Store getStore() {
		return store;
	}

	/**
	 * Key has different subtypes based upon its usage. Currently we support the
	 * String and Number keys. Used in expression that use the =, >, >=, <,
	 * <=, or BETWEEN operators
	 * 
	 * @param k
	 * @throws IOException 
	 */
	public void insert(Key k) {
		int t = getMinimumDegree();

		Node r = root;
		if(r.getNumKeys() == 2*t - 1) {
			Node s = allocateNode();
			setRoot(s);
			s.setLeaf(false);
			s.setNumKeys(0);
			s.setChild(1, r);
			splitChild(s, 1);
			insertNonFull(s, k);
		} else
			insertNonFull(r, k);
	}

	private void insertNonFull(Node x, Key k) {
		int i = x.getNumKeys();
		if(x.isLeaf()) {
			while (i >= 1 && k.compareTo(x.getKey(i)) < 0) {
				x.setKey(i+1, x.getKey(i));
				i = i-1;
			}
			x.setKey(i+1, k);
			x.setNumKeys(x.getNumKeys()+1);
			store.save(x);
		} else {
			while(i >= 1 && k.compareTo(x.getKey(i)) < 0)
				i = i-1;
			i = i+1;
			// x.getChild(i); DISK-READ is performed as part of getChild
			int t = getMinimumDegree();
			if(x.getChild(i).getNumKeys() == 2*t - 1) {
				splitChild(x, i);
				if(k.compareTo(x.getKey(i)) > 0)
					i = i+1;
			}
			insertNonFull(x.getChild(i), k);
		}
	}

	public void delete(Key k) {
		delete(getRoot(), k);
	}

	public void delete(Node x, Key k) {
		int t = getMinimumDegree();

		int i = 1;
		while(i <= x.getNumKeys() && k.compareTo(x.getKey(i)) > 0)
			i = i+1;

		// found key
		if(i <= x.getNumKeys() && k.compareTo(x.getKey(i)) == 0)
		{
			if(x.isLeaf()) {
				// case 1
				x.removeKey(i);
			} else {
				// case 2a
				Node y = x.getChild(i);
				if(y.getNumKeys() >= t) {
					Key kPredecessor = y.getKey(y.getNumKeys());
					delete(y, kPredecessor);
					x.setKey(i, kPredecessor);
				} else {
					Node z = x.getChild(i+1);
					// case 2b
					if(z.getNumKeys() >= t) {
						Key kSuccessor = z.getKey(1);
						delete(z, kSuccessor);
						x.setKey(i, kSuccessor);
					} else { // case 2c
						// remove k from x
						x.removeKey(i);

						mergeNodes(y, k, z);
						markFree(z);
						delete(y, k);
					}
				}
			}
			store.save(x);
		} else { // key not found
			if(!x.isLeaf()) {
				Node y = x.getChild(i);
				if(y.getNumKeys() == t-1) {
					Node leftSibling = (i > 1) ? x.getChild(i-1) : null;
					Node rightSibling = (i <= x.getNumKeys()) ? x.getChild(i+1) : null;

					// case 3a
					if( (leftSibling != null && leftSibling.getNumKeys() >= t) ||
							(rightSibling != null && rightSibling.getNumKeys() >= t) ) {
						if(leftSibling != null && leftSibling.getNumKeys() >= t) {
							// Insert x's i-1th key at the beginning of y
							for(int j = y.getNumKeys(); j >= 1; j--)
								y.setKey(j+1, y.getKey(j));

							if(!y.isLeaf()) {
								for(int j = y.getNumKeys()+1; j >= 1; j--)							
									y.setChild(j+1, y.getChild(j));

								// Move the child pointer from the leftSibling to y
								y.setChild(1, leftSibling.getChild(leftSibling.getNumKeys()+1));
							}

							y.setKey(1, x.getKey(i-1));
							y.setNumKeys(y.getNumKeys()+1);

							// Move the last key in the leftSibling to x
							x.setKey(i-1, leftSibling.getKey(leftSibling.getNumKeys()));
							leftSibling.setNumKeys(leftSibling.getNumKeys()-1);

						} else {
							// Insert x's ith key at the end of y
							y.setKey(y.getNumKeys()+1, x.getKey(i));
							y.setNumKeys(y.getNumKeys()+1);

							// Move the child pointer from the rightSibling to y
							if(!y.isLeaf())
								y.setChild(y.getNumKeys()+1, rightSibling.getChild(1));

							// Move the first key in the rightSibling to x
							x.setKey(i, rightSibling.getKey(1));
							rightSibling.removeKey(1);						
						}
					}
					// case 3b
					else if( (leftSibling == null || leftSibling.getNumKeys() == t-1) &&
							(rightSibling == null || rightSibling.getNumKeys() == t-1) ) {
						if(leftSibling != null) { // push x's i-1th key down into the merged node
							mergeNodes(leftSibling, x.getKey(i-1), y);
							markFree(y);
							x.removeKey(i-1);
							y = leftSibling;
						} else { // push x's ith key down into the merged node
							mergeNodes(y, x.getKey(i), rightSibling);
							markFree(rightSibling);
							x.removeKey(i);
						}
						// remove the ith key and i+1th child pointer from x
						if(x == root && x.getNumKeys() == 0) { // x is the root node
							markFree(x);
							setRoot(y);
						} 
					}
				}  
				delete(y, k);
			}
		}
	}

	/**
	 * This method merges node y with the key k and the node z.
	 * @param y The node that has the key k and the Node z appended/merged with it
	 * @param k
	 * @param z
	 */
	private void mergeNodes(Node y, Key k, Node z) {
		int t = getMinimumDegree();

		assert(y.getNumKeys() == t-1);
		assert(z.getNumKeys() == t-1);

		y.setKey(t, k);

		for(int i = t+1; i < 2*t; i++)
			y.setKey(i, z.getKey(i-t));

		if(!y.isLeaf())
			for(int i = t+1; i <= 2*t; i++)
				y.setChild(i, z.getChild(i-t));
		y.setNumKeys(2*t - 1);
	}

	private void markFree(Node x) {
		// TODO: mark the node x as free
	}

	// The following could have an iterator interface

	/**
	 * A key may match multiple entries, especially if there are duplicate keys
	 * 
	 * @param key
	 * @return
	 */
	public List<Key> getEquals(Key key) {
		return null;
	}

	public List<Key> getBetween(Key startKey, Key endKey) {
		return null;
	}

	public List<Key> getGE(Key key) {
		return null;
	}

	public List<Key> getGT(Key key) {
		return null;
	}

	public List<Key> getLT(Key key) {
		return null;
	}

	public List<Key> getLE(Key key) {
		return null;
	}

	/**
	 * Returns the list of all entries sorted by key
	 * 
	 * @return
	 */
	public List<Key> getAll() {
		return null;
	}

	public IndexEntry search(Key key) {
		return search(root, key);
	}

	// private methods
	private IndexEntry search(Node node, Key key) {
		int i = 1;
		while (i <= node.getNumKeys() && key.compareTo(node.getKey(i)) > 0)
			i = i + 1;

		if (i <= node.getNumKeys() && key.compareTo(node.getKey(i)) == 0)
			return new BTreeEntry(node, i);
		else if (node.isLeaf())
			return null;
		else
			return search(node.getChild(i), key);
	}

	private Node getRoot() {
		if(root == null) {
			Node r = new BTreeNode(store.getRootAddress());		
			store.load(r);
			setRoot(r);	
		}

		return root;
	}

	private Node allocateNode() {
		Node result = new BTreeNode();
		store.load(result);

		return result;
	}

	private void splitChild(Node x, int i) {
		Node z = allocateNode();
		Node y = x.getChild(i);
		z.setLeaf(y.isLeaf());
		int t = getMinimumDegree();

		z.setNumKeys(t - 1);
		for(int j = 1; j < t; j++)
			z.setKey(j, y.getKey(j+t));

		if(!y.isLeaf())
			for(int j = 1; j <= t; j++)
				z.setChild(j, y.getChild(j+t));
		y.setNumKeys(t-1);

		for(int j = x.getNumKeys()+1; j > i; j--)
			x.setChild(j+1, x.getChild(j));
		x.setChild(i+1, z);

		for(int j = x.getNumKeys(); j >= i; j--)
			x.setKey(j+1, x.getKey(j));

		x.setKey(i, y.getKey(t));
		x.setNumKeys(x.getNumKeys()+1);

		store.save(y);
		store.save(z);
		store.save(x);
	}

	// The following 2 methods should be part of an Iterator interface
	private BTreeEntry next(Node x, int i) {
		return null;
	}

	private BTreeEntry previous(Node x, int i) {
		return null;
	}

	/*
	 *   root   - G, M, P, X
	 *   child1 - A, C, D, E
	 *   child2 - J, K
	 *   child3 - N, O
	 *   child4 - R, S, T, U, V
	 *   child5 - Y, Z
	 */
	protected CLRSTest getTestData() {
		root.setAddress(1);
		BTreeNode child1 = new BTreeNode(2);
		BTreeNode child2 = new BTreeNode(3);
		BTreeNode child3 = new BTreeNode(4);
		BTreeNode child4 = new BTreeNode(5);
		BTreeNode child5 = new BTreeNode(6);

		// Let the cache know about these nodes
		store.load(root);
		store.load(child1);
		store.load(child2);
		store.load(child3);
		store.load(child4);
		store.load(child5);

		root.setChild(1, child1);
		root.setChild(2, child2);
		root.setChild(3, child3);
		root.setChild(4, child4);
		root.setChild(5, child5);

		// root
		root.setKey(1, indexType.getKey("G", -1));
		root.setKey(2, indexType.getKey("M", -1));
		root.setKey(3, indexType.getKey("P", -1));
		root.setKey(4, indexType.getKey("X", -1));
		root.setLeaf(false);
		root.setNumKeys(4);

		// child1
		child1.setKey(1, indexType.getKey("A", -1));
		child1.setKey(2, indexType.getKey("C", -1));
		child1.setKey(3, indexType.getKey("D", -1));
		child1.setKey(4, indexType.getKey("E", -1));
		child1.setLeaf(true);
		child1.setNumKeys(4);

		// child2
		child2.setKey(1, indexType.getKey("J", -1));
		child2.setKey(2, indexType.getKey("K", -1));
		child2.setLeaf(true);
		child2.setNumKeys(2);

		// child3
		child3.setKey(1, indexType.getKey("N", -1));
		child3.setKey(2, indexType.getKey("O", -1));
		child3.setLeaf(true);
		child3.setNumKeys(2);

		// child4
		child4.setKey(1, indexType.getKey("R", -1));
		child4.setKey(2, indexType.getKey("S", -1));
		child4.setKey(3, indexType.getKey("T", -1));
		child4.setKey(4, indexType.getKey("U", -1));
		child4.setKey(5, indexType.getKey("V", -1));
		child4.setLeaf(true);
		child4.setNumKeys(5);

		// child5
		child5.setKey(1, indexType.getKey("Y", -1));
		child5.setKey(2, indexType.getKey("Z", -1));
		child5.setLeaf(true);
		child5.setNumKeys(2);

		CLRSTest result = new CLRSTest();
		result.root = root;
		result.child1 = child1;
		result.child2 = child2;
		result.child3 = child3;
		result.child4 = child4;
		result.child5 = child5;

		return result;
	}	

	private class BTreeEntry implements IndexEntry {
		Node node; // Node containing the key
		int index; // The index where the key is present in the node
		Key key;

		public BTreeEntry(Node node, int i) {
			this.node = node;
			this.index = i;
			this.key = node.getKey(i);
		}

		@Override
		public int compareTo(Key o) {
			return key.compareTo(o);
		}

		@Override
		public Node getNode() {
			return node;
		}

		@Override
		public int getIndex() {
			return index;
		}

		@Override
		public Object getName() {
			return key.getName();
		}

		@Override
		public long getData() {
			return key.getData();
		}
	}

	private class BTreeNode implements Node {

		/** The number of keys currently stored in this node **/
		private int numKeys;

		/** The file contents of this node **/
		private byte[] data;

		/**
		 * The offset of this node in the file. This also represents the child
		 * pointer and uniquely identifies the node. What this means, the node
		 * cannot be moved around in the file. This is fine because this
		 * algorithm does not do any file movement.
		 **/
		private long address;

		/** true if this represents a leaf node and has no child pointers **/
		private boolean leaf;		


		public BTreeNode() {
			data = new byte[store.getBlockSize()]; 
		}

		public BTreeNode(long address) {
			this();
			this.address = address; 
		}		

		@Override
		public int getNumKeys() {
			return numKeys;
		}

		public void setNumKeys(int numKeys) {
			this.numKeys = numKeys;
		}

		@Override
		public byte[] getData() {
			return data;
		}

		@Override
		public long getAddress() {
			return address;
		}

		@Override
		public void setAddress(long offset) {
			this.address = offset;
		}

		@Override
		public boolean isLeaf() {
			return leaf;
		}

		@Override
		public void setLeaf(boolean leaf) {
			this.leaf = leaf;
		}

		/**
		 * @param keyIndex
		 *            ranges from 1 to numKeys
		 * @return
		 */
		@Override
		public Key getKey(int keyIndex) {
			int offset = NODE_HEADER_LEN + CHILD_PTR_LEN*keyIndex + (indexType.getSize()*(keyIndex-1));
			return indexType.fromBytes(data, offset);
		}

		@Override
		public void setKey(int keyIndex, Key key) {
			byte[] source = indexType.toBytes(key);
			int destPos = NODE_HEADER_LEN + CHILD_PTR_LEN*keyIndex + (indexType.getSize()*(keyIndex-1));
			System.arraycopy(source, 0, data, destPos, source.length);
		}

		@Override
		public void removeKey(int i) {
			if( (root == this && getNumKeys() == 0) || getNumKeys() == getMinimumDegree()-1 )
				throw new IllegalArgumentException("Cannot remove since minimum number of keys already reached");

			// remove ith key
			if(!isLeaf())
				for(int j = i+1; j <= getNumKeys(); j++)
					setChild(j, getChild(j+1));

			for(int j = i; j < getNumKeys(); j++)
				setKey(j, getKey(j+1));
			setNumKeys(getNumKeys()-1);
		}

		/**
		 * @param pointerIndex
		 *            ranges from 1 to numKeys+1
		 * @return
		 */
		@Override
		public Node getChild(int childIndex) {
			if(childIndex > getNumKeys()+1 || childIndex < 1)
				throw new IllegalArgumentException("Invalid child index - the child index value has to be from 1 and numKeys+1 inclusive");

			int childOffset = NODE_HEADER_LEN + CHILD_PTR_LEN*(childIndex-1) + (indexType.getSize()*(childIndex-1));
			long childAddress = ByteArrayHelper.toLong(data, childOffset);

			if(childAddress == 0)
				return null;

			Node result = store.get(childAddress);
			if(result == null) {
				result = new BTreeNode(childAddress);
				store.load(result);
			}

			return (BTreeNode) result;
		}

		public void setChild(int childIndex, Node child) {
			byte[] source = ByteArrayHelper.toBytes(child.getAddress());
			int destPos = NODE_HEADER_LEN + CHILD_PTR_LEN*(childIndex-1) + (indexType.getSize()*(childIndex-1));
			System.arraycopy(source, 0, data, destPos, source.length);
		}

		@Override
		public void postLoad() {
			this.numKeys = ByteArrayHelper.toInt(data, 0); 
			this.address = ByteArrayHelper.toLong(data, Integer.SIZE/8);

			byte leaf = data[Integer.SIZE/8 + Long.SIZE/8];
			if(leaf != 0)
				this.setLeaf(true);
		}

		@Override
		public void preSave() {
			// save numKeys
			byte[] numKeysBytes = ByteArrayHelper.toBytes(numKeys);
			System.arraycopy(numKeysBytes, 0, data, 0, numKeysBytes.length);

			// save address
			byte[] addressBytes = ByteArrayHelper.toBytes(address);
			System.arraycopy(addressBytes, 0, data, Integer.SIZE/8, addressBytes.length);

			// save isLeaf
			if(isLeaf())
				data[Integer.SIZE/8 + Long.SIZE/8] = 1;
		}

	}
}
