/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.util.btree;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;

import ws.ladder.util.ClassUtil;

public class DiskStore extends AbstractStore {
	/*
	 * Block size
	 * Entry size
	 * Index Type
	 * Root node address
	 * 
	 */
	public static final int HEADER_LENGTH = Integer.SIZE/8 + Integer.SIZE/8 + Byte.SIZE/8 + Long.SIZE/8;
	
	private File             file;
	private RandomAccessFile raf;
	private FileChannel      fileChannel;
	
	public DiskStore(File dataFile, int blockSize, int entrySize, IndexTypeEnum type) throws IOException {
		super(blockSize, entrySize, type);
		
		// If this is an existing dataFile, then validate the blockSize, entrySize and type
		if(type == null)
			throw new RuntimeException("Unsupported index type");
		
		if (blockSize < MIN_BLOCK_SIZE) {
			throw new IllegalArgumentException("block size must be at least "
					+ MIN_BLOCK_SIZE + " bytes");
		}
		if (entrySize <= 0) {
			throw new IllegalArgumentException(
					"value size must be larger than 0");
		}
		if (blockSize < BTree.MIN_NUM_KEYS * entrySize) {
			throw new IllegalArgumentException(
					"Block size is too small, ensure it can store at least "
							+ BTree.MIN_NUM_KEYS + " entries");
		}

		file = dataFile;

		// Make sure the file exists
		file.createNewFile();

		raf = new RandomAccessFile(file, "rw");
		fileChannel = raf.getChannel();

		if (fileChannel.size() == 0L) {
			// Empty file, initialize it with the specified parameters
			this.blockSize = blockSize;
			this.entrySize = entrySize;
			this.type = type;
			this.rootAddress = HEADER_LENGTH;		

			// The first node is the root node
			writeHeader();
		} else {
			// Read parameters from file
			readHeader();
		}

		// Calculate the meta information
		initGlobalMeta();	
	}
	
	private void initGlobalMeta() {

	}	
	
	/**
	 * Closes any opened files and release any resources used by this B-Tree.
	 * Once the B-Tree has been closes, it can no longer be used.
	 **/
	public void close() throws IOException {
		raf.close();

		fileChannel = null;
		raf = null;
		file = null;
	}

	public void flush() throws IOException {
		fileChannel.force(false);
	}

	private void writeHeader() throws IOException {
		ByteBuffer buf = ByteBuffer.allocate(getHeaderSize());
		buf.putInt(blockSize);
		buf.putInt(entrySize);
		buf.put(type.getIndex());
		buf.putLong(rootAddress);

		buf.rewind();

		fileChannel.write(buf, 0L);
	}

	private void readHeader() throws IOException {
		ByteBuffer buf = ByteBuffer.allocate(getHeaderSize());
		fileChannel.read(buf, 0L);

		buf.rewind();

		blockSize = buf.getInt();
		entrySize = buf.getInt();
		type = IndexTypeEnum.getType(buf.get());
		rootAddress = buf.getLong();
	}	
	
	@Override
	public void save(Node x) {
		ByteBuffer buf = ByteBuffer.wrap(x.getData());
		buf.limit(blockSize);
		try {
			if(x.getAddress() == 0) // Node is not persisted - New node
				x.setAddress(fileChannel.size());

			x.preSave();
			fileChannel.write(buf, x.getAddress());
		} catch (IOException e) {
			throw ClassUtil.wrapRun(e);
		}		
	}
	
	@Override
	public void load(Node x) {
		ByteBuffer buf = ByteBuffer.wrap(x.getData());
		buf.limit(blockSize);
		try {
			fileChannel.read(buf, x.getAddress());
			cachedNodes.put(x.getAddress(), x);
			x.postLoad();
		} catch (IOException e) {
			throw ClassUtil.wrapRun(e);
		}		
	}

	@Override
	public int getHeaderSize() {
		return HEADER_LENGTH;
	}

	@Override
	public IndexType getIndexType() {
		if(this.type == IndexTypeEnum.STRING)
			return new StringIndexType(entrySize);
		
		return null;
	}

	@Override
	public void setRoot(Node root) {
		rootAddress = root.getAddress();
		try {
			writeHeader();
		} catch (IOException e) {
			throw ClassUtil.wrapRun(e);
		}
	}
}