/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.util.btree;

import java.io.UnsupportedEncodingException;

import ws.ladder.util.ClassUtil;

public class StringIndexType implements IndexType {
	public static final int MAX_LENGTH = 120;  

	private int    maxLength = MAX_LENGTH; // Needs to be a multiple of 12 for UTF-8

	public StringIndexType() {
		this(MAX_LENGTH);
	}

	public StringIndexType(int value) {
		this.maxLength = value - (getHeaderSize()+getDataLength());
		if(maxLength < 1)
			throw new IllegalArgumentException("maxLength value is too small, ensure that the entrySize is large enough to hold the header and datalength");
	}	

	public int getMaxLength() {
		return maxLength;
	}
	
	private int getHeaderSize() {
		return Short.SIZE/8;
	}
	
	private int getDataLength() {
		return Long.SIZE/8;
	}

	@Override	
	public int getSize() {
		return getHeaderSize() + getMaxLength() + getDataLength();
	}

	@Override
	public int compare(Key key1, Key key2) {
		return key1.compareTo(key2);
	}

	@Override
	public byte[] toBytes(Key k) {
		byte[] result = new byte[getSize()];

		if(!StringKey.class.isAssignableFrom(k.getClass()))
			throw new RuntimeException("Key should represent a String type");

		String val = (String)k.getName();
		byte[] valBytes;
		try {
			valBytes = val.getBytes("UTF-8");
		} catch (UnsupportedEncodingException e) {
			throw ClassUtil.wrapRun(e);
		}

		if(valBytes.length > Short.MAX_VALUE)
			throw new RuntimeException("The key length cannot be greater than " + Short.MAX_VALUE);

		short valLen = (short) valBytes.length;
		byte[] lenBytes = ByteArrayHelper.toBytes(valLen);
		System.arraycopy(lenBytes, 0, result, 0, lenBytes.length);
		System.arraycopy(valBytes, 0, result, getHeaderSize(), valLen);
		System.arraycopy(ByteArrayHelper.toBytes(k.getData()), 0, result, getHeaderSize()+getMaxLength(), getDataLength());

		return result;
	}

	@Override
	public Key fromBytes(byte[] data, int offset) {
		short valLen = ByteArrayHelper.toShort(data, offset);
		
		byte[] temp = new byte[valLen];
		System.arraycopy(data, offset + getHeaderSize(), temp, 0, valLen);		
		
		try {
			String val = new String(temp, "UTF-8");
			
			long dataPointer = ByteArrayHelper.toLong(data, offset + getHeaderSize() + getMaxLength());
			return getKey(val, dataPointer);
		} catch (UnsupportedEncodingException e) {
			throw ClassUtil.wrapRun(e);
		}
	}

	@Override
	public Key getKey(Object value, long data) {
		if(!String.class.isAssignableFrom(value.getClass()))
			throw new RuntimeException("Value should be a String type");
		
		return new StringKey((String) value, data);
	}		

	public static class StringKey implements Key {
		String name;
		long   dataPointer;

		private StringKey(String value, long data) {
			this.name = value;
			this.dataPointer = data;
		}

		@Override
		public Object getName() {
			return name;
		}

		@Override
		public int compareTo(Key key) {
			if(!StringKey.class.isAssignableFrom(key.getClass()))
				throw new RuntimeException("Cannot compare with a non-String key");

			StringKey o = (StringKey) key;		
			return this.name.compareTo((String) o.getName());
		}

		@Override
		public long getData() {
			return dataPointer;
		}
	}

}